﻿using System.Linq;
using Styx.CommonBot;
using Styx.CommonBot.POI;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System;
using System.Collections.Generic;
using CommonBehaviors.Actions;
using Fpsware.Library3.Helpers;
using Fpsware.Library3.Utility;
using Styx;
using Styx.CommonBot.Frames;
using Styx.Pathing;
using Styx.TreeSharp;
using Action = Styx.TreeSharp.Action;
using ActionFail = Fpsware.Library3.Extensions.TreeSharpExt.ActionFail;

namespace Fpsware.Library3.Composites
{
    public static class Harvestable
    {
        public delegate T Selection<out T>(object context);

        private static int _maxDistance = 60;
        /// <summary>
        /// The maximum distance to scan for harvestable items
        /// </summary>
        public static int MaxDistance
        {
            get { return _maxDistance; }
            set { _maxDistance = value; }
        }

        /// <summary>
        /// Return the closest WoWGameObject of the following type Herb, Mineral, Chest
        /// </summary>
        public static WoWGameObject LootableWoWGameObject
        {
            get
            {
                // if our POI is valid return it as the lootable WoWUnit
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Harvest && BotPoi.Current.AsObject != null && BotPoi.Current.AsObject.IsValid && BotPoi.Current.AsObject.ToGameObject().CanLoot)
                {
                    return BotPoi.Current.AsObject.ToGameObject();
                }

                // if our POI is not valid then find one and assign it to the POI
                WoWGameObject lootableObject =
                    ObjectManager.GetObjectsOfType<WoWGameObject>(true)
                        .Where(o => (o.CanLoot) && o.IsValid && (o.IsMineral || o.IsChest || o.IsHerb) && o.CanUse() && !Blacklist.Contains(o, BlacklistFlags.Interact) && o.Distance <= MaxDistance)
                        .OrderBy(o => o.DistanceSqr)
                        .FirstOrDefault();

                // Return our lootabled objectToHarvest
                if (lootableObject != null)
                {
                    if (lootableObject.Location.IsObjectUnderground() || lootableObject.Location.IsObjectInWater())
                    {
                        Log.Info(string.Format("Blacklisted {0} [for harvest] because its " + (lootableObject.Location.IsObjectUnderground() ? "under ground" : "under water"), lootableObject.Name));
                        Blacklist.Add(lootableObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                        return null;
                    }
                    
                    if (!StyxWoW.Me.IsFlying && !Navigator.CanNavigateFully(StyxWoW.Me.Location, lootableObject.Location))
                    {
                        Log.Info(string.Format("Blacklisted {0} [for harvest] because we can't navigate to it", lootableObject.Name));
                        Blacklist.Add(lootableObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                        return null;
                    }
                    if (lootableObject.CanLoot) BotPoi.Current = new BotPoi(lootableObject, PoiType.Harvest);
                    return lootableObject;

                }

                // Clear our POI - just in case its not yet cleared
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Harvest)
                {
                    BotPoi.Clear();
                }

                return null;
            }
        }


        public class QuickHarvestBehaviorPS : PrioritySelector
        {
            public Selection<WoWGameObject> ObjectToHarvestDelegate { get; set; }
            private WoWGameObject CachedLootObject { get; set; }
            private ulong CachedGUID;
            private bool HasCastHerbing = false;
            private bool HasCastMining = false;

            public QuickHarvestBehaviorPS(Selection<WoWGameObject> objectToHarvest)
            {
                ObjectToHarvestDelegate = objectToHarvest;

                Children = CreateChildren();
            }


            private List<Composite> CreateChildren()
            {
                return new List<Composite>()
                {
                    new ActionFail(context =>
                    {
                        CachedLootObject = ObjectToHarvestDelegate(context);
                        CachedGUID = CachedLootObject.Guid;

                    }),

                    new ActionFail(context=> Lua.DoString("SetCVar('AutoLootDefault','1')")),

                    // Stop moving if we're in range
                    new Decorator(ret => StyxWoW.Me.IsMoving && CachedLootObject.Distance <= CachedLootObject.InteractRange * 0.9, new Movement.MoveStop()),

                    // Move to object
                    new Decorator(ret => !StyxWoW.Me.IsFlying && CachedLootObject.Distance > CachedLootObject.InteractRange * 0.95, new Movement.MoveTo(ret=> CachedLootObject.Location)),
                    new Decorator(ret => StyxWoW.Me.IsFlying && CachedLootObject.Distance > CachedLootObject.InteractRange * 0.95, new Movement.MoveTo(ret=> CachedLootObject.Location, ret=> Movement.MovementProviderType.Flightor, ret=> Movement.MountStrategyType.MountFlying, ret=> false)),

                    // Interact     
                    new Decorator(ret=> CachedLootObject != null && CachedLootObject.IsValid && CachedLootObject.Distance < CachedLootObject.InteractRange * 0.99 && !WoWMovement.ActiveMover.IsMoving && LootFrame.Instance != null && CachedLootObject.CanUseNow() && !LootFrame.Instance.IsVisible,
                        new Sequence(
                            
                            // First loot attempt
                            new Action(context => { HasCastHerbing = false; HasCastMining = false; }),
                            new Action(context => { try { CachedGUID = CachedLootObject.Guid; } catch { } }),
                            new Action(context => CachedLootObject.Interact()),
                            new WaitContinue(TimeSpan.FromMilliseconds(500), ret=> false, new ActionAlwaysSucceed()),
                            new DecoratorContinue(ret=> StyxWoW.Me.IsCasting && StyxWoW.Me.CastingSpell.Name == "Mining", new Action(context => HasCastMining = true)),
                            new DecoratorContinue(ret=> StyxWoW.Me.IsCasting && StyxWoW.Me.CastingSpell.Name == "Herbalism", new Action(context => HasCastHerbing = true)),
                            
                            // Wait for loot window to appear and then disappear
                            new WaitContinue(TimeSpan.FromMilliseconds(3000), ret=> LootFrame.Instance != null && LootFrame.Instance.IsVisible, new ActionAlwaysSucceed()),
                            new WaitContinue(TimeSpan.FromMilliseconds(3000), ret=> LootFrame.Instance != null && !LootFrame.Instance.IsVisible, new ActionAlwaysSucceed()),
                            new WaitContinue(TimeSpan.FromMilliseconds(500), ret=> false, new ActionAlwaysSucceed()),
                           
                            // Blacklist to prevent double attempts at harvesting
                            new DecoratorContinue(ret=> HasCastMining || HasCastHerbing, 
                                new Action(context =>
                                {
                                    Blacklist.Add(CachedGUID, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                                    Lua.DoString("SetCVar('AutoLootDefault','0')");
                                })),

                            // Everything looted, loot frame
                            new DecoratorContinue(ret=> LootFrame.Instance != null && LootFrame.Instance.IsVisible && LootFrame.Instance.LootItems <= 0, 
                                new Action(context =>
                                {
                                    Log.Info("==== closing loot frame");
                                    Lua.DoString("CloseLoot()");
                                })),

                            // Finally just toggle off auto loot            
                            new Action(context=> Lua.DoString("SetCVar('AutoLootDefault','0')"))


                        )),

                    new ActionFail(context=> Lua.DoString("SetCVar('AutoLootDefault','0')")),

                };
            }
        }
    }
}
