﻿using System;
using System.Collections.Generic;
using System.Linq;
using CommonBehaviors.Actions;
using Fpsware.Library3.Helpers;
using Fpsware.Library3.Utility;
using Styx;
using Styx.CommonBot;
using Styx.CommonBot.Frames;
using Styx.CommonBot.POI;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
using ActionFail = Fpsware.Library3.Extensions.TreeSharpExt.ActionFail;

namespace Fpsware.Library3.Composites
{
    public static class Loot
    {

        public delegate T Selection<out T>(object context);

        private static int _maxDistance = 60;
        /// <summary>
        /// The maximum distance to scan for lootable object
        /// </summary>
        public static int MaxDistance
        {
            get { return _maxDistance; }
            set { _maxDistance = value; }
        }

        /// <summary>
        /// Return the closest lootable WoWUnit
        /// </summary>
        public static WoWUnit LootableWoWUnit
        {
            get
            {
                // if our POI is valid return it as the lootable WoWUnit
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Loot && BotPoi.Current.AsObject != null && BotPoi.Current.AsObject.IsValid && BotPoi.Current.AsObject.ToUnit().CanLoot && !Blacklist.Contains(BotPoi.Current.AsObject.Guid, BlacklistFlags.Loot))
                {
                    return BotPoi.Current.AsObject.ToUnit();
                }

                // if our POI is not valid then find one and assign it to the POI
                WoWUnit lootableObject =
                    ObjectManager.GetObjectsOfType<WoWUnit>(true)
                        .Where(o => (o.CanLoot) && o.IsValid && !Blacklist.Contains(o, BlacklistFlags.Loot) && o.Distance < MaxDistance) //&& Navigator.CanNavigateFully(StyxWoW.Me.Location, o.Location))
                        .OrderBy(o => o.DistanceSqr)
                        .FirstOrDefault();

                // Return our lootabled unit
                if (lootableObject != null)
                {
                    if (!Navigator.CanNavigateFully(StyxWoW.Me.Location, lootableObject.Location))
                    {
                        Log.Info(string.Format("Blacklisted {0} [for looting] because we can't navigate to it", lootableObject.Name));
                        Blacklist.Add(lootableObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                        Blacklist.Add(lootableObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                        return null;
                    }
                    if (lootableObject.CanLoot) BotPoi.Current = new BotPoi(lootableObject, PoiType.Loot);
                    return lootableObject;
                   
                }

                // Clear our POI - just in case its not yet cleared
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Loot)
                {
                    BotPoi.Clear();
                }
                
                return null;
            }
        }
        public static WoWUnit SkinnableWoWUnit
        {
            get
            {
                // Skinnable unit?
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Skin && BotPoi.Current.AsObject != null && BotPoi.Current.AsObject.IsValid && BotPoi.Current.AsObject.ToUnit().CanSkin)
                {
                    return BotPoi.Current.AsObject.ToUnit();
                }

                // if our POI is not valid then find one and assign it to the POI
                WoWUnit skinnableObject =
                    ObjectManager.GetObjectsOfType<WoWUnit>(true)
                        .Where(o => (o.CanSkin) && !Blacklist.Contains(o, BlacklistFlags.Interact) && !Query.IsInCompetition(o, 15) && o.Distance <= MaxDistance)
                        .OrderBy(o => o.DistanceSqr)
                        .FirstOrDefault();

                // Return our lootabled unit
                if (skinnableObject != null)
                {
                    if (!StyxWoW.Me.IsFlying && !Navigator.CanNavigateFully(StyxWoW.Me.Location, skinnableObject.Location))
                    {
                        Log.Info(string.Format("Blacklisted {0} [for skinning] because we can't navigate to it", skinnableObject.Name));
                        Blacklist.Add(skinnableObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                        Blacklist.Add(skinnableObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                        return null;
                    }

                    BotPoi.Current = new BotPoi(skinnableObject, PoiType.Skin);
                    return skinnableObject;

                }

                // Clear our POI - just in case its not yet cleared
                /*
                if (BotPoi.Current != null && BotPoi.Current.Type == PoiType.Skin)
                {
                    BotPoi.Clear();
                }
                 */

                return null;
            }
        }


        public class LootingBehaviorPS : PrioritySelector
        {
            public Selection<WoWUnit> LootableUnitDelegate { get; set; }
            public Selection<bool> SkinTheMobDelegate { get; set; }
            //public Selection<WoWObject> LootableObjectDelegate { get; set; }

            private WoWUnit CachedLootObject { get; set; }
            private bool SkinTheMob = false;
            private int LootableType = 0; // 0 = WoWUnit, 1 = WoWObject (herb, ore)

            public LootingBehaviorPS(Selection<WoWUnit> lootableUnit, Selection<bool> skinMob = null)
            {
                LootableUnitDelegate = lootableUnit;
                LootableType = 0;
                SkinTheMobDelegate = skinMob ?? (context=> false);

                Children = CreateChildren();
            }

            /*
            public LootingBehaviorPS(Selection<WoWObject> lootableObject)
            {
                LootableObjectDelegate = lootableObject;
                LootableType = 1;

                Children = CreateChildren();
            }
             */

            private int _lootAttempts = 0;
            private int _interactAttempts = 0;

            

            private List<Composite> CreateChildren()
            {
                return new List<Composite>()
                {
                    new ActionFail(context =>
                    {
                        CachedLootObject = LootableUnitDelegate(context);
                        SkinTheMob = SkinTheMobDelegate(context);
                    }),


                    // Stop moving if we're in range
                    new Decorator(ret => CachedLootObject.Distance <= CachedLootObject.InteractRange * 0.9 && WoWMovement.ActiveMover.IsMoving, new Movement.MoveStop()),

                    // Move to object
                    new Decorator(ret => CachedLootObject.Distance > CachedLootObject.InteractRange * 0.95, new Movement.MoveTo(ret=> CachedLootObject.Location)),


                    /*
                    new ThrottlePasses(1,2, new PrioritySelector(
                        new Action(context => { Log.Info("==== testing: " + CachedLootObject.Name + " can skin: " + CachedLootObject.CanSkin); })
                        
                        )),
                     */
                    //new TreeSharpExt.LogMessage("==== 1 ok....CachedLootObject.Name: " + CachedLootObject.Name),
                    
                    // Update the Cached object and reset the number of attempts
                    /*
                    new ThrottlePasses(1, 5,
                        new PrioritySelector(
                            
                            new Decorator(ret=> SkinTheMob && (CachedLootObject == null || !CachedLootObject.CanSkin),
                                new Action(context =>
                                {
                                    Log.FailLog("=== skinnable cached item no longer valid");
                                    if (CachedLootObject != null) Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                                    _lootAttempts = 0;
                                    _interactAttempts = 0;
                                })),
                            
                            new Decorator(ret=> !SkinTheMob && (CachedLootObject == null || !CachedLootObject.CanLoot),
                                new Action(context =>
                                {
                                    Log.FailLog("=== cached item no longer valid");
                                    if (CachedLootObject != null) Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                                    _lootAttempts = 0;
                                    _interactAttempts = 0;
                                }))
                        )
                    ),
                     */

                    //new TreeSharpExt.LogMessage("==== 2......"),
                     
                    // If the loot window is empty close it
                    new Decorator(ret=> LootFrame.Instance.IsVisible && LootFrame.Instance.LootItems <= 0, 
                        new Action(context =>
                        {
                            Log.Info("=== nothing in loot window, lets close it.");
                            Lua.DoString("CloseLoot()");
                            Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                            _lootAttempts = 0;
                            _interactAttempts = 0;

                        })),

                        //new TreeSharpExt.LogMessage("==== 3"),

                    // After 10 attempts close the loot window and try again
                    new Decorator(ret=> _lootAttempts == 10 && LootFrame.Instance != null && LootFrame.Instance.IsVisible && LootFrame.Instance.LootItems >= 1, 
                        new Action(context =>
                        {
                            Log.Info("=== looted 10 times but this time we'll close the window and try again");
                            //Blacklist.Add(Lootable.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(5));
                            Lua.DoString("CloseLoot()");
                            _lootAttempts += 1;

                        })),


                        //new TreeSharpExt.LogMessage("==== 4"),
                    // After X attempts blacklist it and give up
                    new Decorator(ret=> _lootAttempts >= 20 || _interactAttempts >= 7, 
                        new Action(context =>
                        {
                            if (_lootAttempts >= 21) Log.Info("=== looted 20 times but still no luck. Blacklisting now");
                            if (_interactAttempts >= 5) Log.Info("=== unable to loot after 5 attempts. Blacklisting now");
                            Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.All, TimeSpan.FromMinutes(5));
                            Lua.DoString("CloseLoot()");
                            _lootAttempts = 0;
                            _interactAttempts = 0;

                        })),
                        
                        //new TreeSharpExt.LogMessage("==== 4"),
                    // Do the actual looting
                    //new ThrottlePasses(TimeSpan.FromMilliseconds(1200), 
                        new Decorator(ret=> LootFrame.Instance != null && LootFrame.Instance.IsVisible && LootFrame.Instance.LootItems >= 1,// && _lootStopwatch.ElapsedMilliseconds >= 1200, 
                            new Sequence(
                            new Action(context =>
                                {
                                    //_lootStopwatch.Reset();
                                    _lootAttempts += 1;

                                    Log.Info("=== LOOT ATTEMPT: " + _lootAttempts);
                                
                                    for (int i = 0; i < LootFrame.Instance.LootItems; i++)
                                    {
                                        if (StyxWoW.Me.Combat)// || !LootFrame.Instance.IsVisible)
                                        {
                                            Lua.DoString("CloseLoot()");
                                            return RunStatus.Failure;
                                        }
                                        LootSlotInfo lootInfo = LootFrame.Instance.LootInfo(i);
                                        Log.Info("=== Looting: " + lootInfo.LootName);
                                        LootFrame.Instance.Loot(i);
                                    }

                                    if (LootFrame.Instance.LootItems <= 0)
                                    {
                                        Log.Info("==== we looted everything from the loot window, blacklisting for LOOT");
                                        Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(5));
                                        _lootAttempts = 0;
                                        _interactAttempts = 0;
                                    }
                                    return RunStatus.Success;

                                }),

                            new WaitContinue(1, ret=> false, new ActionAlwaysSucceed()),

                            // Do it once more
                            new Action(context =>
                                {
                                    for (int i = 0; i < LootFrame.Instance.LootItems; i++)
                                    {
                                        if (StyxWoW.Me.Combat)
                                        {
                                            Lua.DoString("CloseLoot()");
                                            return RunStatus.Failure;
                                        }
                                        LootSlotInfo lootInfo = LootFrame.Instance.LootInfo(i);
                                        Log.Info("=== Looting: " + lootInfo.LootName);
                                        LootFrame.Instance.Loot(i);
                                    }

                                    if (LootFrame.Instance.LootItems <= 0)
                                    {
                                        Log.Info("==== we looted everything from the loot window second time around");
                                        Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(5));
                                        _lootAttempts = 0;
                                        _interactAttempts = 0;
                                        Lua.DoString("CloseLoot()");
                                    }
                                    return RunStatus.Success;

                                }),
                                new WaitContinue(1, ret=> false, new ActionAlwaysSucceed()),
                                
                                // Loot now
                                new DecoratorContinue(ret=> SkinTheMob && !LootFrame.Instance.IsVisible, 
                                    new Sequence(
                                        new Action(context =>
                                        {
                                            Log.Info("===== first skin attempt: " + _interactAttempts);
                                            //_interactAttempts += 1;
                                            CachedLootObject.Interact();
                                        }),
                                        new WaitContinue(TimeSpan.FromMilliseconds(2000), ret=> false, new ActionAlwaysSucceed()),
                                        new WaitContinue(5, ret=> !StyxWoW.Me.IsCasting, new ActionAlwaysSucceed()),

                                        new DecoratorContinue(ret=> LootFrame.Instance.IsVisible, 
                                            new Sequence(
                                                
                                                new Action(context =>
                                                {
                                                    //_lootStopwatch.Reset();
                                                    _lootAttempts += 1;

                                                    Log.Info("=== GET ALL LOOT: " + _lootAttempts);
                                
                                                    for (int i = 0; i < LootFrame.Instance.LootItems; i++)
                                                    {
                                                        if (StyxWoW.Me.Combat)// || !LootFrame.Instance.IsVisible)
                                                        {
                                                            Lua.DoString("CloseLoot()");
                                                            return RunStatus.Failure;
                                                        }
                                                        LootSlotInfo lootInfo = LootFrame.Instance.LootInfo(i);
                                                        Log.Info("=== getting Looting: " + lootInfo.LootName);
                                                        LootFrame.Instance.Loot(i);
                                                    }

                                                    if (LootFrame.Instance.LootItems <= 0)
                                                    {
                                                        Log.Info("==== we looted all skins from the loot window");
                                                        Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(5));
                                                        _lootAttempts = 0;
                                                        _interactAttempts = 0;
                                                    }
                                                    return RunStatus.Success;

                                                }
                                                
                                                
                                                ))
                                        
                                        
                                        
                                        ))
                                        )

                            )
                    ),

                    
                    // Skin the mob    
                    new Decorator(ret=> SkinTheMob && CachedLootObject != null && CachedLootObject.CanSkin && CachedLootObject.Distance < CachedLootObject.InteractRange * 0.99 && !WoWMovement.ActiveMover.IsMoving && LootFrame.Instance != null && !LootFrame.Instance.IsVisible,// && _lootStopwatch.ElapsedMilliseconds >= 1000,
                        new Sequence(
                            new Action(context =>
                            {
                                Log.Info("===== InteractAttempts: " + _interactAttempts);
                                _interactAttempts += 1;
                                CachedLootObject.Interact();
                            }),
                            new WaitContinue(TimeSpan.FromSeconds(2), ret=> false, new ActionAlwaysSucceed()),
                            new WaitContinue(5, ret=> !StyxWoW.Me.IsCasting, new ActionAlwaysSucceed())
                        )),


                    // Interact with the object
                    new Decorator(ret=> !SkinTheMob && CachedLootObject != null && CachedLootObject.CanLoot && CachedLootObject.Distance < CachedLootObject.InteractRange * 0.99 && !WoWMovement.ActiveMover.IsMoving && LootFrame.Instance != null && !LootFrame.Instance.IsVisible,// && _lootStopwatch.ElapsedMilliseconds >= 1000,
                        new Sequence(
                            new Action(context =>
                            {
                                Log.Info("===== InteractAttempts: " + _interactAttempts);
                                _interactAttempts += 1;
                                CachedLootObject.Interact();
                            }),
                            new WaitContinue(TimeSpan.FromMilliseconds(1200), ret=> false, new ActionAlwaysSucceed()),
                            new WaitContinue(5, ret=> !StyxWoW.Me.IsCasting, new ActionAlwaysSucceed())
                        ))
                        
                };
            }
        }

        public class QuickLootingBehaviorPS : PrioritySelector
        {
            public Selection<WoWUnit> UnitDelegate { get; set; }
            public Selection<bool> SkinTheMobDelegate { get; set; }

            private WoWUnit CachedLootObject { get; set; }
            private bool SkinTheMob = false;
            private bool HasCastSkinning = false;
            private ulong CachedGUID;

            public QuickLootingBehaviorPS(Selection<WoWUnit> unit, Selection<bool> skinMob)
            {
                UnitDelegate = unit;
                SkinTheMobDelegate = skinMob ?? (context => false);

                Children = CreateChildren();
            }


            private List<Composite> CreateChildren()
            {
                return new List<Composite>()
                {
                    new ActionFail(context =>
                    {
                        //CachedGUID = CachedLootObject.Guid;
                        SkinTheMob = SkinTheMobDelegate(context);
                        CachedLootObject = UnitDelegate(context);

                    }),

                    new ActionFail(context=> Lua.DoString("SetCVar('AutoLootDefault','1')")),

                    // Stop moving if we're in range
                    new Decorator(ret => StyxWoW.Me.IsMoving && CachedLootObject.Distance <= CachedLootObject.InteractRange * 0.9, new Movement.MoveStop()),

                    // Move to object
                    new Decorator(ret => !StyxWoW.Me.IsFlying && CachedLootObject.Distance > CachedLootObject.InteractRange * 0.95, new Movement.MoveTo(ret=> CachedLootObject.Location)),
                    new Decorator(ret => StyxWoW.Me.IsFlying && CachedLootObject.Distance > CachedLootObject.InteractRange * 0.95, new Movement.MoveTo(ret=> CachedLootObject.Location, ret=> Movement.MovementProviderType.Flightor, ret=> Movement.MountStrategyType.MountFlying, ret=> false)),

                    // Interact     
                    new Decorator(ret=> CachedLootObject != null && CachedLootObject.IsValid && CachedLootObject.Distance < CachedLootObject.InteractRange * 0.99 && !WoWMovement.ActiveMover.IsMoving && LootFrame.Instance != null && !LootFrame.Instance.IsVisible,
                        new Sequence(
                            
                            // First loot attempt
                            new Action(context => HasCastSkinning = false),
                            new Action(context => { try { CachedGUID = CachedLootObject.Guid; } catch { } }),
                            new Action(context => CachedLootObject.Interact()),
                            new WaitContinue(TimeSpan.FromMilliseconds(500), ret=> false, new ActionAlwaysSucceed()),
                            new DecoratorContinue(ret=> StyxWoW.Me.IsCasting && StyxWoW.Me.CastingSpell.Name == "Skinning", new Action(context => HasCastSkinning = true)),
                            
                            // We are or were casting skinning
                            new DecoratorContinue(ret=> HasCastSkinning, 
                                new Sequence(
                                    //new Action(context => { try { CachedGUID = CachedLootObject.Guid; } catch (Exception e) { } }),
                                    new WaitContinue(5, ret=> !StyxWoW.Me.IsCasting, new ActionAlwaysSucceed()),
                                    new WaitContinue(1, ret=> LootFrame.Instance.IsVisible, new ActionAlwaysSucceed()),
                                    //new WaitContinue(2, ret=> false, new ActionAlwaysSucceed()),
                                    new Action(context =>
                                    {
                                        Blacklist.Add(CachedGUID, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                                        Blacklist.Add(CachedGUID, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                                    })
                                    
                                )),

                            // We cast skinning, lets just wait a little bit longer
                            new DecoratorContinue(ret=> HasCastSkinning, new WaitContinue(2, ret=> false, new ActionAlwaysSucceed())),
                            
                            // We didn't cast skinning? Must have just looted the mob
                            new DecoratorContinue(ret=> !HasCastSkinning, 
                                new Sequence(
                                    new WaitContinue(TimeSpan.FromMilliseconds(3000), ret=> LootFrame.Instance != null && !LootFrame.Instance.IsVisible, new ActionAlwaysSucceed())
                                    )),
                            
                            // Everything looted, loot frame
                            new DecoratorContinue(ret=> LootFrame.Instance != null && LootFrame.Instance.IsVisible && LootFrame.Instance.LootItems <= 0, 
                                new Action(context =>
                                {
                                    Log.Info("==== closing loot frame");
                                    Lua.DoString("CloseLoot()");
                                })),

                            // Loot once more - skinning after looting
                            new DecoratorContinue(ret=> SkinTheMob && !HasCastSkinning && CachedLootObject != null && CachedLootObject.IsValid, 
                                new Sequence(
                                    new WaitContinue(TimeSpan.FromMilliseconds(1000), ret=> false, new ActionAlwaysSucceed()),
                                    new WaitContinue(TimeSpan.FromMilliseconds(1500), ret=> CachedLootObject != null && CachedLootObject.IsValid && CachedLootObject.CanSkin, new ActionAlwaysSucceed()),
                                    
                                    // The cached unit it no longer valid bail out here
                                    new Decorator(ret=> CachedLootObject != null && CachedLootObject.IsValid, new Action(context => CachedLootObject.Interact())),
                                    new WaitContinue(TimeSpan.FromMilliseconds(500), ret=> false, new ActionAlwaysSucceed()),
                                    new DecoratorContinue(ret=> StyxWoW.Me.IsCasting && StyxWoW.Me.CastingSpell.Name == "Skinning", new Action(context => HasCastSkinning = true)),
                                    new WaitContinue(5, ret=> !StyxWoW.Me.IsCasting, new ActionAlwaysSucceed()),
                                    new WaitContinue(TimeSpan.FromMilliseconds(1000), ret=> !HasCastSkinning, new ActionAlwaysSucceed()),
                                    new WaitContinue(TimeSpan.FromMilliseconds(3000), ret=> LootFrame.Instance != null && !LootFrame.Instance.IsVisible, new ActionAlwaysSucceed()),
                                    new WaitContinue(TimeSpan.FromMilliseconds(1000), ret=> !HasCastSkinning, new ActionAlwaysSucceed()),
                                    new DecoratorContinue(ret=> HasCastSkinning, 
                                        new Action(context =>
                                        {
                                            Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10));
                                            Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Interact, TimeSpan.FromMinutes(10));
                                            //Log.Info("=== skinning was cast and mob blacklisted");
                                            Lua.DoString("SetCVar('AutoLootDefault','0')");
                                        }))
                                        
                                    )),

                            // Prevent form attempting to double loot
                            new Action(context=> Blacklist.Add(CachedLootObject.Guid, BlacklistFlags.Loot, TimeSpan.FromMinutes(10))),
                            
                            // Finally just toggle off auto loot            
                            new Action(context=> Lua.DoString("SetCVar('AutoLootDefault','0')"))


                        )),

                    new ActionFail(context=> Lua.DoString("SetCVar('AutoLootDefault','0')")),

                };
            }
        }
    }


    

}
