﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using CommonBehaviors.Actions;
using Fpsware.Library3.Extensions;
using Fpsware.Library3.Helpers;
using Fpsware.Library3.Utility;
using JetBrains.Annotations;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace Fpsware.Botbases.RAFBot2
{
    public static class EncounterLogicActions
    {
        private static bool IsMobTargetingValid(this EncounterLogic e, WoWUnit u)
        {
            return e.SourceTarget == SourceTarget.IsTargetingMe && u.IsTargetingMe() ||
                   e.SourceTarget == SourceTarget.IsMyCurrentTarget && StyxWoW.Me.GotTarget &&
                   u.Guid == StyxWoW.Me.CurrentTargetGuid || e.SourceTarget == SourceTarget.AnyHostileNearBy;
        }

        private static bool IsMobCastingEncounterLogic(this WoWUnit u, IEnumerable<EncounterLogic> logic, out EncounterLogic matchedEncounterLogic)
        {
            matchedEncounterLogic = null;

            foreach (EncounterLogic e in logic.Where(e => u.Entry == e.MobID && u.CastingSpellId == e.SpellID && u.Distance <= e.EffectiveDistance && e.IsMobTargetingValid(u) && u.IsSafelyFacing(StyxWoW.Me, e.Degrees)))
            {
                matchedEncounterLogic = e;
                return true;
            }

            return false;
        }

        private static bool IsMobAuraEncounterLogic(this WoWUnit u, IEnumerable<EncounterLogic> logic, out EncounterLogic matchedEncounterLogic)
        {
            matchedEncounterLogic = null;

            foreach (EncounterLogic e in logic.Where(e => u.Entry == e.MobID && u.Auras.Any(a => a.Value.SpellId == e.SpellID) && u.Distance <= e.EffectiveDistance && e.IsMobTargetingValid(u)))// && u.IsSafelyFacing(StyxWoW.Me, e.Degrees)))
            {
                matchedEncounterLogic = e;
                return true;
            }

            return false;
        }

        private static bool IsMobPresentEncounterLogic(this WoWUnit u, IEnumerable<EncounterLogic> logic, out EncounterLogic matchedEncounterLogic)
        {
            matchedEncounterLogic = null;

            foreach (EncounterLogic e in logic.Where(e => u.Entry == e.MobID && u.Distance <= e.EffectiveDistance && e.IsMobTargetingValid(u)))
            {
                matchedEncounterLogic = e;
                return true;
            }

            return false;
        }

        private static bool IsEncounterLogicMet(out EncounterLogic matchedEncounterLogic, out WoWUnit triggerUnit)
        {
            EncounterLogic encounterLogic = null;

            #region Query Casting logic first

            IEnumerable<EncounterLogic> castingLogicsList = from logic in EncounterLogicsList() where logic.QueryType == QueryType.SpellCast select logic;

            if (castingLogicsList.Any())
            {
                IEnumerable<WoWUnit> castingUnits = from unit in ObjectManager.GetObjectsOfTypeFast<WoWUnit>() where unit.IsHostile && unit.IsCasting && unit.DistanceSqr < 60*60 select unit;
                if (castingUnits.Any())
                {

                    WoWUnit validUnit = (from unit in castingUnits where unit.IsMobCastingEncounterLogic(castingLogicsList, out encounterLogic) select unit).FirstOrDefault();

                    triggerUnit = validUnit;
                    matchedEncounterLogic = encounterLogic;
                    //return validUnit != null;
                    if (validUnit != null) return true;
                }
            }

            #endregion

            #region Query Aura logic second

            IEnumerable<EncounterLogic> auraLogicsList = from logic in EncounterLogicsList() where logic.QueryType == QueryType.Aura select logic;
            if (auraLogicsList.Any())
            {
                // Initial list of units to query
                IEnumerable<WoWUnit> unitsToCheck = from unit in ObjectManager.GetObjectsOfTypeFast<WoWUnit>() where unit.IsHostile && unit.DistanceSqr < 60*60 select unit;

                // Find the first unit matching our encounter logic
                WoWUnit validUnit = (from unit in unitsToCheck where unit.IsMobAuraEncounterLogic(auraLogicsList, out encounterLogic) select unit).FirstOrDefault();

                triggerUnit = validUnit;
                matchedEncounterLogic = encounterLogic;
                if (validUnit != null) return true;
            }

            #endregion

            #region Query Mob avoid logic lastly

            IEnumerable<EncounterLogic> mobAvoidLogicsList = from logic in EncounterLogicsList() where logic.QueryType == QueryType.Mob select logic;
            if (mobAvoidLogicsList.Any())
            {
                // Initial list of units to query
                IEnumerable<WoWUnit> unitsToCheck = from unit in ObjectManager.GetObjectsOfTypeFast<WoWUnit>() where unit.DistanceSqr < 60 * 60 select unit;

                // Find the first unit matching our encounter logic
                WoWUnit validUnit = (from unit in unitsToCheck where unit.IsMobPresentEncounterLogic(mobAvoidLogicsList, out encounterLogic) select unit).FirstOrDefault();

                triggerUnit = validUnit;
                matchedEncounterLogic = encounterLogic;
                if (validUnit != null) return true;
            }

            #endregion


            // Nothing at all found
            matchedEncounterLogic = null;
            triggerUnit = null;
            return false;
        }

        private static EncounterLogic _triggeredEncounterLogic = new EncounterLogic();
        private static WoWUnit _triggeredUnit = null;


        /// <summary>
        /// Is the triggered encounter logic still in progress? While we may be out of harms way, we still need to check if the bad stuff is still happening.
        /// </summary>
        public static bool IsTriggeredLogicStillInProgress
        {
            get
            {
                if (_triggeredEncounterLogic == null || _triggeredUnit == null || !_triggeredUnit.IsValid || _triggeredUnit.IsDead)
                {
                    _triggeredEncounterLogic = null;
                    _triggeredUnit = null;
                }

                if (_triggeredEncounterLogic == null || _triggeredUnit == null) return false;


                switch (_triggeredEncounterLogic.QueryType)
                {
                    case QueryType.SpellCast:
                        if (_triggeredUnit.IsCasting && _triggeredUnit.CastingSpellId == _triggeredEncounterLogic.SpellID) return true;
                        break;
                    
                    case QueryType.Aura:
                        if (_triggeredUnit.Auras.Any(a => a.Value.SpellId == _triggeredEncounterLogic.SpellID)) return true;
                        break;
                    
                    case QueryType.Mob:
                        break;
                
                    default:
                        return false;
                }

                
                return false;
            }
        }

        public static bool IsEncounterConditionMet
        {
            get
            {
                //if (_triggeredEncounterLogic.)

                // If our encounter logic list is empty then don't do anything 
                if (EncounterLogicsList().Count <= 0)
                {
                    _triggeredEncounterLogic = null;
                    return false;
                }

                EncounterLogic encounterLogic = null;
                WoWUnit triggerUnit = null;
                
                // Has encounter logic been triggered?
                if (!IsEncounterLogicMet(out encounterLogic, out triggerUnit))
                {
                    //_triggeredEncounterLogic = null;
                    return false;
                }

                // Assign our encounter results and the unit that triggered our result
                EncounterResults.TriggerUnit = triggerUnit;
                EncounterResults.Action = encounterLogic.Action;
                EncounterResults.ActionDistance = encounterLogic.ActionDistance;
                EncounterResults.ActionName = encounterLogic.ActionName;

                _triggeredEncounterLogic = encounterLogic;
                _triggeredUnit = triggerUnit;

                return true;
            }
        }

        private static bool IsEncounterConditionMet_OLD
        {
            get
            {
                if (EncounterLogicsList().Count <= 0) return false;

                // Check all spell casting logics first. Get a list of encounter logic that check for spell casting
                //List<EncounterLogic> castingLogicsList = EncounterLogicsList().Where(l => l.QueryType == QueryType.SpellCast).ToList();
                IEnumerable<EncounterLogic> castingLogicsList = from l in EncounterLogicsList() where l.QueryType == QueryType.SpellCast select l;

                if (castingLogicsList.Any())
                {
                    //List<WoWUnit> castingUnits = ObjectManager.GetObjectsOfTypeFast<WoWUnit>() .Where( u => u.IsHostile && u.IsCasting && u.Distance < 40) //&& .ToList();
                    /*
                    IEnumerable<WoWUnit> castingUnits = from u in ObjectManager.GetObjectsOfTypeFast<WoWUnit>()
                        where u.IsHostile && u.IsCasting && u.DistanceSqr < 60*60
                        select u;
                     */

                    IEnumerable<WoWUnit> castingUnits = from u in ObjectManager.GetObjectsOfTypeFast<WoWUnit>()
                                                        where u.IsHostile && u.IsCasting && u.DistanceSqr < 60 * 60
                                                        select u;


                    // Get list of units that initially meet our criteria
                    //List<WoWUnit> castingUnits = ObjectManager.GetObjectsOfType<WoWUnit>(true, false)
                    //List<WoWUnit> castingUnits = ObjectManager.GetObjectsOfTypeFast<WoWUnit>() .Where( u => u.IsHostile && u.IsCasting && u.Distance < 40) //&& .ToList();

                    if (castingUnits.Any())
                    {

                        //WoWUnit validUnit = (from u in castingUnits where u.IsMobCastingEncounterLogic(castingLogicsList) select u).FirstOrDefault();



                        // Now find the first exact match
                        foreach (EncounterLogic e in castingLogicsList)
                        {
                            WoWUnit unit =
                                castingUnits.FirstOrDefault(
                                    u =>
                                        u.Entry == e.MobID && u.CastingSpellId == e.SpellID &&
                                        u.Distance <= e.EffectiveDistance &&

                                        (e.SourceTarget == SourceTarget.IsTargetingMe && u.IsTargetingMe() ||
                                         e.SourceTarget == SourceTarget.IsMyCurrentTarget && StyxWoW.Me.GotTarget &&
                                         u.Guid == StyxWoW.Me.CurrentTargetGuid ||
                                         e.SourceTarget == SourceTarget.AnyHostileNearBy) &&

                                        u.IsSafelyFacing(StyxWoW.Me, e.Degrees));
                            if (unit == null) continue;

                            // got one!
                            EncounterResults.TriggerUnit = unit;
                            EncounterResults.Action = e.Action;
                            EncounterResults.ActionDistance = e.ActionDistance;
                            EncounterResults.ActionName = e.ActionName;


                            return true;
                        }
                    }
                }

                // Enumerate all logic
                foreach (EncounterLogic logic in EncounterLogicsList())
                {
                    WoWUnit unit = null;
                    switch (logic.QueryType)
                    {
                        case QueryType.Aura:
                            //unit = ObjectManager.GetObjectsOfType<WoWUnit>(true, false) .FirstOrDefault( u => u.Entry == logic.MobID && u.Distance <= logic.EffectiveDistance && u.Auras.Any(a => a.Value.SpellId == logic.SpellID));
                            unit = ObjectManager.GetObjectsOfTypeFast<WoWUnit>().FirstOrDefault(u => u.Entry == logic.MobID && u.Distance <= logic.EffectiveDistance && u.Auras.Any(a => a.Value.SpellId == logic.SpellID));
                            break;

                        case QueryType.Mob:
                            //unit = ObjectManager.GetObjectsOfType<WoWUnit>(true, false) .FirstOrDefault(u => u.Entry == logic.MobID && u.Distance <= logic.EffectiveDistance);
                            unit = ObjectManager.GetObjectsOfTypeFast<WoWUnit>().FirstOrDefault(u => u.Entry == logic.MobID && u.Distance <= logic.EffectiveDistance);
                            break;

                        // CheckAgaint.Casting is handled elsewhere
                    }

                    if (unit != null)
                    {
                        EncounterResults.TriggerUnit = unit;
                        EncounterResults.Action = logic.Action;
                        EncounterResults.ActionDistance = logic.ActionDistance;
                        EncounterResults.ActionName = logic.ActionName;
                        return true;
                    }
                }

                return false;
            }
        }


        private static WoWItem _useItem;
        public static Composite CreateEncounterLogicActions()
        {
            const double interactRangeModifier = 0.99;

            return new PrioritySelector(

                #region Cast Spell
                new Decorator(ret => EncounterResults.Action == Action.CastSpell && SpellManager.HasSpell(EncounterResults.ActionName) && !StyxWoW.Me.IsCasting && !SpellManager.GlobalCooldown && SpellManager.CanCast(EncounterResults.ActionName),
                    new Styx.TreeSharp.Action(context => SpellManager.Cast(EncounterResults.ActionName))),

                #endregion

                #region Use Item
                new Decorator(ret => EncounterResults.Action == Action.UseItem,
                    new Sequence(
                        new Styx.TreeSharp.Action(context =>
                        {
                            _useItem = StyxWoW.Me.CarriedItems.FirstOrDefault(i => i.Name == EncounterResults.ActionName);
                            if (_useItem != null && _useItem.Usable) { _useItem.Use(); }

                        })
                    )),

                #endregion

                #region Move Behind Caster

                new Decorator(ret => EncounterResults.Action == Action.MoveBehindCaster && !Navigator.AtLocation(Movement.CalculatePointBehindTarget(EncounterResults.TriggerUnit, (float)(EncounterResults.TriggerUnit.MeleeDistance() * interactRangeModifier))) && Navigator.CanNavigateFully(StyxWoW.Me.Location, Movement.CalculatePointBehindTarget(EncounterResults.TriggerUnit, (float)(EncounterResults.TriggerUnit.MeleeDistance() * interactRangeModifier))),
                    new Sequence(
                        new WaitContinue(5, ret => !StyxWoW.Me.Stunned, new ActionAlwaysSucceed()),
                        new Styx.TreeSharp.Action(context => Navigator.MoveTo(Movement.CalculatePointBehindTarget(EncounterResults.TriggerUnit, (float)(EncounterResults.TriggerUnit.MeleeDistance() * interactRangeModifier)))),
                        new WaitContinue(5, ret => !StyxWoW.Me.IsMoving || Navigator.AtLocation(Movement.CalculatePointBehindTarget(EncounterResults.TriggerUnit, (float)(EncounterResults.TriggerUnit.MeleeDistance() * interactRangeModifier))), new ActionAlwaysSucceed()),
                        new DecoratorContinue(ret=> !EncounterResults.TriggerUnit.MeIsSafelyBehind,
                            new Styx.TreeSharp.Action(context => Navigator.MoveTo(Movement.CalculatePointBehindTarget(EncounterResults.TriggerUnit, (float)(EncounterResults.TriggerUnit.MeleeDistance() * interactRangeModifier)))))

                        )),

                #endregion

                #region Move to Leader
                new Decorator( ret => EncounterResults.Action == Action.MoveToLeader && RAFGroup.Leader != null && RaFHelper.Leader.Distance > EncounterResults.ActionDistance,
                    new Styx.TreeSharp.Action(context => Navigator.MoveTo(RaFHelper.Leader.Location))
                    ),

                #endregion

                #region Run to Distance
                new Decorator( ret => EncounterResults.Action == Action.RunToDistance && EncounterResults.TriggerUnit.Distance < EncounterResults.ActionDistance,
                    new Sequence(
                        new DecoratorContinue(ret=> StyxWoW.Me.IsMoving && EncounterResults.TriggerUnit.Distance >= EncounterResults.ActionDistance + 2, new Movement.MoveStop() ),
                        Movement.MoveToSafeSpot(ret => EncounterResults.TriggerUnit, ret => 30, ret => EncounterResults.ActionDistance, ret => EncounterResults.ActionDistance + 2, false)
                        )  
                    ),

                #endregion

                #region Stop Casting

                    // Does not stop casting, but should prevent going any further and casting combat spells.
                // Should still allow for healing and such to be cast
                new Decorator(ret => EncounterResults.Action == Action.StopCasting, new ActionAlwaysSucceed())
                #endregion


                );
        }


        /// <summary>
        /// If IsEncounterConditionMet is true the results of the required action and trigger unit will be placed here
        /// </summary>
        public static class EncounterResults 
        {
            public static WoWUnit TriggerUnit { get; set; }
            public static Action Action { get; set; }
            public static string ActionName { get; set; }
            public static int ActionDistance { get; set; }
        }


        [UsedImplicitly]
        private static List<EncounterLogic> _encounterLogics = new List<EncounterLogic>();

        private static string _parsedEncounterLogicString;

        /// <summary>
        /// Return all encounter logics relevant to the current mob entry
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static List<EncounterLogic> EncounterLogicsList(this WoWUnit unit)
        {
            return EncounterLogicsList().Where(e => e.MobID == unit.Entry).ToList();
        }
        
        public static List<EncounterLogic> EncounterLogicsList()
        {
            // Nothing has changed? Return the cached list
            if (string.IsNullOrEmpty(GlobalSetting.Instance.EncounterLogic) || _parsedEncounterLogicString == GlobalSetting.Instance.EncounterLogic)
            {
                return _encounterLogics;
            }

            // Make the list
            _encounterLogics.Clear();
            string[] split = GlobalSetting.Instance.EncounterLogic.Split('|');
            foreach (EncounterLogic encounterLogic in split.Select(s => new EncounterLogic(s)))
            {
                _encounterLogics.Add(encounterLogic);
            }

            _parsedEncounterLogicString = GlobalSetting.Instance.EncounterLogic;
            return _encounterLogics;
        }


        public enum QueryType
        {
            SpellCast = 0,
            Aura,
            Mob
        }

        public enum Action
        {
            RunToDistance = 0,
            MoveBehindCaster,
            StopCasting,
            MoveToLeader,
            CastSpell,
            UseItem
        }

        public enum SourceTarget
        {
            IsMyCurrentTarget = 0,
            IsTargetingMe,
            AnyHostileNearBy
        }
        public class EncounterLogic
        {
            public EncounterLogic()
            {
                Name = null;
                QueryType = QueryType.SpellCast;
                MobID = 0;
                SpellID = 0;
                EffectiveDistance = 0;
                Degrees = 360;
                ActionDistance = 15;
                SourceTarget = SourceTarget.IsMyCurrentTarget;
            }

            public EncounterLogic(string rawString)
            {
                // Assign default values. If an property is missing it will still have a value and not null
                Name = null;
                QueryType = QueryType.SpellCast;
                MobID = 0;
                SpellID = 0;
                EffectiveDistance = 0;
                Degrees = 360;
                ActionDistance = 15;
                SourceTarget = SourceTarget.IsMyCurrentTarget;

                EncounterLogicRawString = rawString;
            }

            public string Name { get; set; }
            public QueryType QueryType { get; set; }
            public SourceTarget SourceTarget{ get; set; }
            public uint SpellID { get; set; }
            public int Degrees { get; set; }
            public uint MobID { get; set; }
            /// <summary>
            /// Effective distance / range of the mob / spell
            /// </summary>
            public int EffectiveDistance { get; set; }
            public Action Action { get; set; }
            public string ActionName { get; set; }
            /// <summary>
            /// The distance used in the action. Typically run away from a mob to this distance or move to X yards of the leader
            /// </summary>
            public int ActionDistance { get; set; }
            


            private string _rawString = null;
            public string EncounterLogicRawString
            {
                get
                {
                    StringBuilder sb = new StringBuilder();
                    switch (QueryType)
                    {
                        case QueryType.Aura:
                        // Bad Aura [Source:Aura,SpellID:213658,MobID:4254,Distance:15]
                        sb.AppendFormat("{0} [CheckAgainst:Aura,SourceTarget:{4},SpellID:{1},MobID:{2},Distance:{3}]", Name, SpellID, MobID, EffectiveDistance, SourceTarget);
                        break;

                        case QueryType.Mob:
                        //Scary Ghost [Source:Mob,MobID:1485,Distance:15]
                        sb.AppendFormat("{0} [CheckAgainst:Mob,SourceTarget:{3},MobID:{1},Distance:{2}]", Name, MobID, EffectiveDistance, SourceTarget);
                        break;

                        case QueryType.SpellCast:
                        //Hurty Fire Spell [Source:SpellCast,SpellID:213658,MobID:4254,Distance:15,Radiud:145]
                        sb.AppendFormat("{0} [CheckAgainst:SpellCast,SourceTarget:{5},SpellID:{1},MobID:{2},Degrees:{3},Distance:{4}]", Name, SpellID, MobID, Degrees, EffectiveDistance, SourceTarget);
                        break;
                    }

                    switch (Action)
                    {
                        case Action.CastSpell:
                            sb.AppendFormat(" Action[CastSpell:{0}]", ActionName);
                            break;

                        case Action.UseItem:
                            sb.AppendFormat(" Action[UseItem:{0}]", ActionName);
                            break;

                        case Action.MoveToLeader:
                            sb.AppendFormat(" Action[MoveToLeader:{0}]",ActionDistance);
                            break;

                        case Action.RunToDistance:
                            sb.AppendFormat(" Action[RunToDistance:{0}]", ActionDistance);
                            break;

                        case Action.MoveBehindCaster:
                            sb.AppendFormat(" Action[MoveBehindCaster]");
                            break;

                        case Action.StopCasting:
                            sb.AppendFormat(" Action[StopCasting]");
                            break;
                    }

                    return sb.ToString();
                }

                set
                {
                    _rawString = value;
                    string name, spellId, mobId, distance, degrees, actionDistance, sourceTargetString;

                    // Find our source
                    //string checkAgainstString = Regex.Match(_rawString, @"(?<=CheckAgainst\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value.Trim();
                    string checkAgainstString = GetRXProperty(_rawString, @"(?<=CheckAgainst\:).*?(?=[\],])", "SpellCast");
                    QueryType queryType;

                    // Try to parse the string as the proper enum, otherwise use the default of SpellCast
                    if (!Enum.TryParse(checkAgainstString, out queryType))
                    {
                        Log.FailLog("===== danger danger danger..... SOURCE tryParse failed");
                        Log.FailLog("=== RawString: " + _rawString);
                        queryType = QueryType.SpellCast;
                    }


                    sourceTargetString = GetRXProperty(_rawString, @"(?<=SourceTarget\:).*?(?=[\],])", SourceTarget.IsMyCurrentTarget.ToString());
                    SourceTarget sourceTarget;
                    if (!Enum.TryParse(sourceTargetString, out sourceTarget))
                    {
                        Log.FailLog("===== danger danger danger..... SourceTarget tryParse failed");
                        Log.FailLog("=== RawString: " + _rawString);
                        sourceTarget = SourceTarget.IsMyCurrentTarget;
                    }
                    SourceTarget = sourceTarget;

                    QueryType = queryType;
                    switch (queryType)
                    {
                        case QueryType.Aura:
                            //Bad Aura [Source:Aura,SpellID:213658,MobID:4254,Distance:15]
                            QueryType = QueryType.Aura;
                            /*
                            name = Regex.Match(_rawString, @"(?<=).*?(?=\[)", RegexOptions.IgnoreCase).Value.Trim();
                            spellId = Regex.Match(_rawString, @"(?<=SpellID\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            mobId = Regex.Match(_rawString, @"(?<=MobID\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            distance = Regex.Match(_rawString, @"(?<=,Distance\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                             */
                            Name = GetRXProperty(_rawString, @"(?<=).*?(?=\[)", "");
                            SpellID = (uint)GetRXProperty(_rawString, @"(?<=SpellID\:).*?(?=[\],])", 0);
                            MobID = (uint)GetRXProperty(_rawString, @"(?<=MobID\:).*?(?=[\],])", 0);
                            EffectiveDistance = (int)GetRXProperty(_rawString, @"(?<=,Distance\:).*?(?=[\],])", 0);
                            

                            //Name = name;
                            //SpellID = Convert.ToUInt32(spellId);
                            //MobID = Convert.ToUInt32(mobId);
                            //EffectiveDistance = Convert.ToUInt16(distance);
                            Degrees = 0;

                            break;

                        case QueryType.Mob:
                            //Scary Ghost [Source:Mob,MobID:1485,Distance:15]
                            QueryType = QueryType.Mob;
                            name = Regex.Match(_rawString, @"(?<=).*?(?=\[)", RegexOptions.IgnoreCase).Value.Trim();
                            mobId = Regex.Match(_rawString, @"(?<=MobID\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            distance = Regex.Match(_rawString, @"(?<=,Distance\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;

                            Name = name;
                            MobID = Convert.ToUInt32(mobId);
                            EffectiveDistance = Convert.ToUInt16(distance);
                            SpellID = 0;
                            Degrees = 0;

                            break;

                        case QueryType.SpellCast:
                            //Hurty Fire Spell [Source:SpellCast,SpellID:213658,MobID:4254,Degrees:145,Distance:15]
                            QueryType = QueryType.SpellCast;
                            name = Regex.Match(_rawString, @"(?<=).*?(?=\[)", RegexOptions.IgnoreCase).Value.Trim();
                            spellId = Regex.Match(_rawString, @"(?<=SpellID\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            mobId = Regex.Match(_rawString, @"(?<=MobID\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            degrees = Regex.Match(_rawString, @"(?<=Degrees\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;
                            distance = Regex.Match(_rawString, @"(?<=,Distance\:).*?(?=[\],])", RegexOptions.IgnoreCase).Value;

                            Name = name;
                            SpellID = Convert.ToUInt32(spellId);
                            MobID = Convert.ToUInt32(mobId);
                            EffectiveDistance = Convert.ToUInt16(distance);
                            Degrees = Convert.ToUInt16(degrees);
                            break;   
                    }


                    string actionString = Regex.Match(_rawString, @"(?<=Action\[).*?(?=[\]\:])", RegexOptions.IgnoreCase).Value.Trim();
                    Action action;

                    // Try to parse the string as the proper enum, otherwise use the default of SpellCast
                    if (!Enum.TryParse(actionString, out action))
                    {
                        Log.FailLog("===== danger danger danger..... ACTION tryParse failed");
                        Log.FailLog("=== RawString: " + _rawString);
                        action = Action.MoveBehindCaster;
                    }
                    
                    Action = action;
                    switch (action)
                    {
                        case Action.UseItem:
                        case Action.CastSpell:
                            ActionName = Regex.Match(_rawString, @"(?<=Action\[CastSpell:|Action\[UseItem:).*?(?=[\]\:,])", RegexOptions.IgnoreCase).Value.Trim();
                            break;

                        case Action.RunToDistance:
                            actionDistance = Regex.Match(_rawString, @"(?<=Action\[RunToDistance:).*?(?=[\]\:,])", RegexOptions.IgnoreCase).Value;
                            ActionDistance = Convert.ToUInt16(actionDistance);
                            break;
                        
                        case Action.MoveToLeader:
                            actionDistance = Regex.Match(_rawString, @"(?<=Action\[MoveToLeader:).*?(?=[\]\:,])", RegexOptions.IgnoreCase).Value;
                            ActionDistance = Convert.ToUInt16(actionDistance);
                            break;

                        //case Action.MoveBehindCaster:
                        //case Action.StopCasting:
                        default:
                            ActionName = "";
                            break;
                    }


                }
            }

        }

        private static string GetRXProperty(string source, string regEx, string defaultOnFailure = "")
        {
            string result = Regex.Match(source, regEx, RegexOptions.IgnoreCase).Value.Trim();
            if (!string.IsNullOrEmpty(result)) return result;

            Log.FailLog(string.Format("EncounterLogic Failure: in GetRXProperty, Source:\"{0}\" RegEx:{1}", source, regEx));
            return defaultOnFailure;
        }

        private static ulong GetRXProperty(string source, string regEx, ulong defaultOnFailure = 0)
        {
            string result = Regex.Match(source, regEx, RegexOptions.IgnoreCase).Value.Trim();
            if (!string.IsNullOrEmpty(result))
            {
                ulong resultUlong = Convert.ToUInt32(result);
                return resultUlong;
            }

            Log.FailLog(string.Format("EncounterLogic Failure: in GetRXProperty, Source:\"{0}\" RegEx:{1}", source, regEx));
            return defaultOnFailure;
            
        }

    }

}
