﻿// Originally contributed by Chinajade.
//
// LICENSE:
// This work is licensed under the
//     Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.
// also known as CC-BY-NC-SA.  To view a copy of this license, visit
//      http://creativecommons.org/licenses/by-nc-sa/3.0/
// or send a letter to
//      Creative Commons // 171 Second Street, Suite 300 // San Francisco, California, 94105, USA.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Windows.Media;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Action = Styx.TreeSharp.Action;

namespace Fpsware.Library3.Extensions
{
    public class TreeSharpExt
    {
        public class ActionFail : Composite
        {
            public ActionFail() { }

            public ActionFail(ActionDelegate action)
            {
                Runner = action;
            }

            public ActionFail(ActionSucceedDelegate action)
            {
                FailRunner = action;
            }

            public ActionDelegate Runner { get; private set; }
            public ActionSucceedDelegate FailRunner { get; private set; }

            protected virtual RunStatus Run(object context)
            {
                return RunStatus.Failure;
            }

            protected RunStatus RunAction(object context)
            {
                if (Runner != null) return Runner(context);

                if (FailRunner != null)
                {
                    FailRunner(context);
                    return RunStatus.Failure;
                }
                RunStatus runStatus = Run(context);

                return runStatus;
            }

            protected override sealed IEnumerable<RunStatus> Execute(object context)
            {
                while ((LastStatus = RunAction(context)) == RunStatus.Running)
                    yield return RunStatus.Running;

                yield return LastStatus.Value;
            }
        }

        public class DecoratorThrottle : Decorator
        {
            public DecoratorThrottle(TimeSpan throttleTime, Composite composite) : this(context => true, throttleTime, composite)
            {
                // empty
            }


            public DecoratorThrottle(CanRunDecoratorDelegate predicateDelegate, TimeSpan throttleTime, Composite composite) : base(composite)
            {
                _predicateDelegate = predicateDelegate;
                _throttleTime = throttleTime;
            }


            protected override bool CanRun(object context)
            {
                if (_predicateDelegate(context))
                {
                    bool canRun = IsThrottleFinished;

                    if (IsThrottleFinished)
                    { _throttle.Restart(); }

                    return canRun;
                }

                if (_throttle.IsRunning)
                { _throttle.Stop(); }

                return false;
            }

            private bool IsThrottleFinished
            {
                get { return !_throttle.IsRunning || (_throttle.Elapsed > _throttleTime); }
            }

            private readonly CanRunDecoratorDelegate _predicateDelegate;
            private readonly Stopwatch _throttle = new Stopwatch();
            private readonly TimeSpan _throttleTime;
        }

        public class DecoratorContinueThrottle : DecoratorContinue
        {
            public DecoratorContinueThrottle(TimeSpan throttleTime, Composite composite) : this(context => true, throttleTime, composite)
            {
                // empty
            }

            public DecoratorContinueThrottle(CanRunDecoratorDelegate predicateDelegate, TimeSpan throttleTime, Composite composite) : base(composite)
            {
                _predicateDelegate = predicateDelegate;
                _throttleTime = throttleTime;
            }

            protected override bool CanRun(object context)
            {
                if (_predicateDelegate(context))
                {
                    bool canRun = IsFinished;

                    if (IsFinished)
                    {
                        _throttle.Restart();
                    }

                    return canRun;
                }

                if (_throttle.IsRunning)
                {
                    _throttle.Stop();
                }

                return false;
            }

            private bool IsFinished
            {
                get { return !_throttle.IsRunning || (_throttle.Elapsed > _throttleTime); }
            }

            private readonly CanRunDecoratorDelegate _predicateDelegate;
            private readonly Stopwatch _throttle = new Stopwatch();
            private readonly TimeSpan _throttleTime;
        }

        public class LogMessage : Action
        {
            private const string MessagePrefix = "";

            private readonly string _message;
            private readonly bool _displayHbStatus;
            private readonly bool _displayWoWChat;
            private readonly object[] args;

            public LogMessage(string message)
            {
                _message = message;
            }

            protected override RunStatus Run(object context)
            {
                Logging.Write(Color.FromRgb(255, 206, 0), string.Format("{0}{1}{2}", MessagePrefix, _message, args));

                if (_displayHbStatus && TreeRoot.IsRunning)
                {
                    TreeRoot.StatusText = _message;
                    TreeRoot.GoalText = _message;
                }

                if (_displayWoWChat && TreeRoot.IsRunning)
                {
                    Lua.DoString(string.Format(@"print('\124cFF15E61C {0}')", _message));
                }

                if (Parent is Selector) return RunStatus.Failure;
                return RunStatus.Success;
            }
        }

        public class LogDebugMessage : Action
        {
            private const string MessagePrefix = "==== DEBUG: ";

            private readonly string _message;

            public LogDebugMessage(string message) { _message = message; }

            protected override RunStatus Run(object context)
            {
                Logging.Write(Color.FromRgb(255, 0, 253), string.Format("{0}{1}", MessagePrefix, _message));
                
                if (Parent is Selector) return RunStatus.Failure;
                return RunStatus.Success;
            }
        }

        public class FrameLockSelector : PrioritySelector
        {
            public FrameLockSelector(params Composite[] children) : base(children)
            {
                // empty
            }


            public FrameLockSelector(ContextChangeHandler contextChange, params Composite[] children) : base(contextChange, children)
            {
                // empty
            }


            public override RunStatus Tick(object context)
            {
                using (StyxWoW.Memory.AcquireFrame())
                {
                    return base.Tick(context);
                }
            }
        }
    }
}
