﻿using System;
using System.Collections.Generic;
using System.Linq;
using Styx;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

// This class was taken from Superbad Druid by Handnavi. All credit to him for this work. 
// I've only changed a few variables and added some addition comments.

namespace Fpsware.Library3.Helpers
{
    /// <summary>
    /// TTL is 'Time To Live'. This class is used to determine how long a mob has until it dies. 
    /// </summary>
    public static class HostileTTL
    {
        private static readonly Dictionary<ulong, DpsInfo> DpsInfos = new Dictionary<ulong, DpsInfo>();
        public static bool DpsMeterInitialized;

        /// <summary>
        ///     Starts the DpsMeter.
        /// </summary>
        public static void Initialize()
        {
            if (DpsMeterInitialized) return;
            DpsInfos.Clear();
            DpsMeterInitialized = true;
        }

        public static void Shutdown()
        {
            if (!DpsMeterInitialized) return;
            DpsInfos.Clear();
            DpsMeterInitialized = false;
        }

        /// <summary>
        /// Pulse the combat information on all mobs in combat. Ideally placed in the CC Pulse routine
        /// </summary>
        public static void Pulse()
        {
            List<WoWUnit> availableUnits = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(u => !u.IsDead && u.Combat && u.DistanceSqr < 50 * 50 && (u.IsTargetingMeOrPet || u.IsTargetingMyPartyMember || u.IsTargetingMyRaidMember || StyxWoW.Me.CurrentTarget == u)).ToList();

            foreach (WoWUnit unit in availableUnits)
            {
                if (!DpsInfos.ContainsKey(unit.Guid))
                {
                    DpsInfo di = new DpsInfo
                    {
                        Unit = unit,
                        CombatStart = ConvDate2Timestam(DateTime.Now),
                        StartHealth = unit.CurrentHealth,
                        StartHealthMax = unit.MaxHealth
                    };
                    DpsInfos.Add(unit.Guid, di);
                }
                else
                {
                    DpsInfo di = DpsInfos[unit.Guid];

                    uint currentLife = unit.CurrentHealth;
                    int currentTime = ConvDate2Timestam(DateTime.Now);
                    int timeDiff = currentTime - di.CombatStart;
                    uint hpDiff = di.StartHealth - currentLife;

                    if (hpDiff > 0)
                    {
                        long fullTime = timeDiff * di.StartHealthMax / hpDiff;
                        long pastFirstTime = (di.StartHealthMax - di.StartHealth) * timeDiff / hpDiff;
                        long calcTime = di.CombatStart - pastFirstTime + fullTime - currentTime;
                        if (calcTime < 1) calcTime = 1;

                        //calc_time is a int value for time to die (seconds) so there's no need to do SecondsToTime(calc_time)
                        long timeToDie = calcTime;
                        di.CombatTimeLeft = new TimeSpan(0, 0, (int)timeToDie);
                    }

                    // The unit has more health now than when we last checked. It was healed or something.
                    if (hpDiff <= 0)
                    {
                        //unit was healed,resetting the initial values
                        // So now we update all the appropriate values and start over
                        di.StartHealth = unit.CurrentHealth;
                        di.StartHealthMax = unit.MaxHealth;
                        di.CombatStart = ConvDate2Timestam(DateTime.Now);

                        //Lets do a little trick and calculate with seconds / u know Timestamp from unix? we'll do so too
                        di.CombatTimeLeft = unit.IsCritter || unit.MaxHealth <= StyxWoW.Me.MaxHealth / 4
                            ? new TimeSpan(0, 0, 5)
                            : new TimeSpan(0, 0, 16);
                    }

                    // The unit's current health is at its max health? WTF
                    if (currentLife == di.StartHealthMax)
                    {
                        di.CombatTimeLeft = unit.IsCritter || unit.MaxHealth <= StyxWoW.Me.MaxHealth / 4
                            ? new TimeSpan(0, 0, 5)
                            : new TimeSpan(0, 0, 16);
                    }


                    // .NET makes a copy of the struct when we grab it out of the collection.
                    // Make sure we put the updated version back in!
                    DpsInfos[unit.Guid] = di;
                    KeyValuePair<ulong, DpsInfo>[] removeUnits = DpsInfos.Where(kv => !kv.Value.Unit.IsValid).ToArray();
                    foreach (var t in removeUnits)
                    {
                        DpsInfos.Remove(t.Key);
                    }
                }
            }
        }

        private static int ConvDate2Timestam(DateTime time)
        {
            {
                DateTime date1 = new DateTime(1970, 1, 1);
                DateTime date2 = time;
                TimeSpan ts = new TimeSpan(date2.Ticks - date1.Ticks);
                return (Convert.ToInt32(ts.TotalSeconds));
            }
        }

        /// <summary>
        ///     Returns the estimated combat time left for this unit. (Time until death)
        ///     If the unit is invalid; TimeSpan.MinValue is returned.
        /// </summary>
        /// <param name="u"></param>
        /// <returns></returns>
        public static TimeSpan GetCombatTimeLeft(WoWUnit u)
        {
            if (u == null || !u.IsValid) return new TimeSpan(0, 0, 61);

            //If our current target is a dummy, we set a fixed value for combat time left (61 seconds).
            if (u.Name.Contains("Dummy")) return new TimeSpan(0, 0, 61);
            return DpsInfos.ContainsKey(u.Guid) ? DpsInfos[u.Guid].CombatTimeLeft : TimeSpan.MinValue;
        }

        /// <summary>
        /// Wrapper for GetCombatTimeLeft(WoWUnit u)
        /// </summary>
        /// <param name="u"></param>
        /// <returns></returns>
        public static double CombatSecondsLeft(this WoWUnit u)
        {
            //Log.DebugLog("=== Combat Seconds left: " + GetCombatTimeLeft(u).TotalSeconds);
            return GetCombatTimeLeft(u).TotalSeconds;
        }

        /// <summary>
        /// Returns TRUE if the estimated combat seconds remaining is equal to or greater than 'minCombatSeconds'
        /// </summary>
        /// <param name="u"></param>
        /// <param name="minCombatSeconds"></param>
        /// <returns></returns>
        public static bool IsCombatSecOK(this WoWUnit u, int minCombatSeconds)
        {
            return GetCombatTimeLeft(u).TotalSeconds >= minCombatSeconds;
        }

        #region Nested type: DpsInfo

        private struct DpsInfo
        {
            public int CombatStart;
            public TimeSpan CombatTimeLeft;
            public uint StartHealth;
            public uint StartHealthMax;
            public WoWUnit Unit;
        }

        #endregion
    }
}
