﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using Fpsware.Library3.Extensions;
using Styx;
using System;
using Styx;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.POI;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace Fpsware.Library3.Helpers
{
    public static class Query
    {
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static Stopwatch _latencyTimer = new Stopwatch();
        private static int _latencyCached = 0;

        public static bool IsThisMyComputer(string computernames)
        {
            string[] computerName = computernames.Split('|');
            string currentComputerName = System.Security.Principal.WindowsIdentity.GetCurrent().Name.ToUpper();

            return computerName.Any(s => currentComputerName.ToUpper().Contains(s.Trim().ToUpper()));
        }
        
        /// <summary>
        /// Return WoW latency in milliseconds. If checked before an elapsed time of 30 seconds it will return a fixed value of 300. After 60 seconds has expired the true WoW latency value is cached and returned
        /// </summary>
        public static int WoWLatency
        {
            get
            {
                if (_latencyCached > 0) return _latencyCached;

                if (!_latencyTimer.IsRunning)
                {
                    _latencyTimer.Start();
                    return 300;
                }

                if (_latencyCached == 0 || _latencyTimer.Elapsed.TotalSeconds < 60)
                {
                    return 300;
                }

                if (_latencyTimer.Elapsed.TotalSeconds >= 60)
                {
                    _latencyCached = (int) StyxWoW.WoWClient.Latency;
                    _latencyTimer.Reset();
                    _latencyTimer.Stop();

                    return _latencyCached;
                }

                // If all else fails simply return 300 [ms]
                return 300;
            }
            
        }


        public static bool IsValidForPulling(WoWUnit wowUnit, double nonCompeteDistance = 15)
        {
            return IsValidTarget(wowUnit)
                    && !wowUnit.IsBlacklistedForPulling()
                    //&& IsStateMatch_IgnoreMobsInBlackspots(wowUnit, ignoreMobsInBlackspots)
                    && !IsInCompetition(wowUnit, nonCompeteDistance);
        }

        public static bool IsValidTarget(this WoWObject p)
        {
            string invalidReason = null;

            if (p == null) { invalidReason = "unit == null"; }

            WoWUnit wowUnit = p as WoWUnit;
            if (wowUnit == null)
            {
                //Log.Info("===== !IsValidTarget : " + invalidReason);
                invalidReason = "unit is not WoWUnit";
                return false;
            }
            

            if (!wowUnit.CanSelect) { invalidReason = "!unit.CanSelect"; }
            if (string.IsNullOrEmpty(invalidReason) && !wowUnit.Attackable) { invalidReason = "!unit.Attackable"; }
            if (string.IsNullOrEmpty(invalidReason) && wowUnit.IsFriendly) { invalidReason = "unit.IsFriendly"; }
            if (string.IsNullOrEmpty(invalidReason) && wowUnit.IsDead) { invalidReason = "unit.IsDead"; }
            if (string.IsNullOrEmpty(invalidReason) && Blacklist.Contains(p, BlacklistFlags.All)) { invalidReason = "unit.IsBlacklisted"; }
            if (string.IsNullOrEmpty(invalidReason) && (wowUnit.IsNonCombatPet || wowUnit.IsCritter)) { invalidReason = "unit.IsCritter"; }

            if (!string.IsNullOrEmpty(invalidReason))
            {
                //Log.Info("===== !IsValidTarget : " + invalidReason);

                return false;
            }

            if (wowUnit.Name.Contains("Training Dummy")) return true;
            if (wowUnit.IsHostile) return true;

            return true;
        }





        public static bool IsInCompetition(WoWObject wowObject, double nonCompeteDistance)
        {
            //if (!IsValidTarget(wowObject)) { return false; }

            // Shared world resources are never in competition...
            if (IsSharedWorldResource(wowObject)) { return false; }

            // If unit is tagged by someone else and in combat, it is in competition...
            // N.B. There are some cases where NPCs are shown as tagged by - 
            // someone else yet nobody is actively engaged with said unit. 
            // If unit is not in combat with anyone and it is tagged and nobody else is around it can not be in competition. 
            // On the other hand if we choose to ignore the tagged unit and it's the only NPC that exists in the world then bot will be stuck
            WoWUnit wowUnit = wowObject as WoWUnit;
            bool isTaggedByOther = ((wowUnit != null) && !wowUnit.IsUntagged() && wowUnit.Combat);
            if (isTaggedByOther) { return true; }


            //ProvideBoolDelegate excludeGroupMembers = (potentialGroupMember =>
            Types.Selection<bool> excludeGroupMembers = (potentialGroupMember =>
            {
                WoWPlayer asWoWPlayer = potentialGroupMember as WoWPlayer;
                return (asWoWPlayer != null) && !asWoWPlayer.IsInMyParty;
            });

            bool isPlayersNearby = FindPlayersNearby(wowObject.Location, nonCompeteDistance, excludeGroupMembers).Any();
            bool isCompetitionTimerRunning = InCompetitionTimers.ContainsKey(wowObject.Guid);

            // If players are clear, and competition timer is running...
            // We no longer need the competition timer.
            if (!isPlayersNearby && isCompetitionTimerRunning) { InCompetitionTimers.Remove(wowObject.Guid); }

            // If players are nearby, and we haven't established competition timer...
            // We need to record time at which we start the wait.
            if (isPlayersNearby && !isCompetitionTimerRunning)
            {
                // Add new entry...
                InCompetitionTimers.Add(wowObject.Guid, DateTime.Now);

                // Time to sweep away old 'in competition' entries?
                if ((_inCompetitionSweepTimer == null) || _inCompetitionSweepTimer.IsFinished)
                {
                    // Remove expired 'in competition' entries...
                    DateTime now = DateTime.Now;
                    ulong[] keysToRemove =
                        (from kvp in InCompetitionTimers
                            where now > (kvp.Value + InCompetitionSweepTime)
                            select kvp.Key)
                            .ToArray();

                    foreach (ulong key in keysToRemove) { InCompetitionTimers.Remove(key); }

                    // Reset sweep timer (creating it, if necessary)...
                    if (_inCompetitionSweepTimer == null) { _inCompetitionSweepTimer = new WaitTimer(InCompetitionSweepTime); }

                    _inCompetitionSweepTimer.Reset();
                }
            }


            // If we've been waiting on object too long, it is no longer 'in competition'...
            // NB: Since group membership affects 'nearby players', we must make this test
            // after the competition timers have been updated due to nearby players.
            DateTime waitStart;
            if (InCompetitionTimers.TryGetValue(wowObject.Guid, out waitStart))
            {
                if (DateTime.Now > (waitStart + InCompetitionMaxWaitTime)) { return false; }

                // Fall through, if we haven't been waiting too long...
            }

            return (isPlayersNearby);
        }
        private static readonly Dictionary<ulong, DateTime> InCompetitionTimers = new Dictionary<ulong, DateTime>();
        private static readonly TimeSpan InCompetitionMaxWaitTime = TimeSpan.FromSeconds(90);
        private static readonly TimeSpan InCompetitionSweepTime = TimeSpan.FromSeconds(10/*mins*/ * 60 /*secs*/);
        private static WaitTimer _inCompetitionSweepTimer = null;


        public static IEnumerable<WoWPlayer> FindPlayersNearby(WoWPoint location, double radius, Types.Selection<bool> extraQualifiers = null)
        {
            extraQualifiers = extraQualifiers ?? (wowPlayerContext => true);

            return from wowPlayer in ObjectManager.GetObjectsOfType<WoWPlayer>(true, false)
                where
                    wowPlayer.IsValid
                    && wowPlayer.IsAlive
                    && (wowPlayer.Location.Distance(location) < radius)
                    && extraQualifiers(wowPlayer)
                select wowPlayer;
        }

        public static bool IsSharedWorldResource(WoWObject wowObject)
        {
            bool isSharedResource = false;
            var wowGameObject = wowObject.ToGameObject();
            var wowUnit = wowObject as WoWUnit;

            isSharedResource |= (wowGameObject != null) && SharedGameObjectTypes.Contains(wowGameObject.SubType);
            isSharedResource |= (wowUnit != null) && wowUnit.TappedByAllThreatLists;
            isSharedResource |= (wowUnit != null) && !wowUnit.IsHostile
                                && (wowUnit.IsAnyTrainer
                                    || wowUnit.IsAnyVendor
                                    || wowUnit.IsAuctioneer
                                    || wowUnit.IsBanker
                                    || wowUnit.IsFlightMaster
                                    //|| wowUnit.IsGuard
                                    || wowUnit.IsGuildBanker
                                    || wowUnit.IsInnkeeper
                                    || wowUnit.IsQuestGiver
                                    || wowUnit.IsStableMaster
                                   );

            return isSharedResource;
        }
        private static readonly WoWGameObjectType[] SharedGameObjectTypes =
        {
            WoWGameObjectType.Binder,           // sets hearthstone
            // NOT: WoWGameObjectType.Door,
            // Although it would be nice to include doors, NPCs locked in cages that require a key to open
            // are "door" subtypes.  If we walked up and took one of these resources while it was in competition
            // that would not be good.  Thus, we exclude it from our list of 'shared resources'.
            WoWGameObjectType.GuildBank,
            WoWGameObjectType.Mailbox,
            WoWGameObjectType.MeetingStone,
            WoWGameObjectType.QuestGiver,
            WoWGameObjectType.SpellCaster,      // portals
            WoWGameObjectType.Transport
        };

    }
}
