﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;
using System.Text.RegularExpressions;
using System.ComponentModel;
using System.Threading;
using System.Windows.Media;
using Styx.Common;

namespace Fpsware.Library3.Helpers
{

    public class SvnUpdater
    {
        [Category("Data"), Description("This is the project name in the SVN URL provided by your repository link.\n\nIf using Assembla:\nSVN URL:https://subversion.assembla.com/svn/fpsware-raid-bot/\nThe project name is 'fpsware-raid-bot'.\n\nIf using Google:\nSVN URL:https://fpsware-raidbot.googlecode.com/svn/trunk/\nThe project name is 'fpsware-raidbot'."), DisplayName("Project Name")]
        public string SVNProject { get; set; }

        private string _revisionFile;
        [Category("Data"), Description("Full path and file name of the created revision file. If this file does not exist it will be created."), DisplayName("Revision File")]
        public string RevisionFile
        {
            get
            {
                if (string.IsNullOrEmpty(_revisionFile) && !string.IsNullOrEmpty(DownloadPath))
                {
                    _revisionFile = Path.Combine(DownloadPath, "Revision.xml");
                }

                return _revisionFile;
            }
            set { _revisionFile = value; }
        }

        private string _svnUrl;
        [Category("Data"), Description("The full URL of the SVN to be downloaded from (Google or Assembla). Eg http://your-project-name.googlecode.com/svn/trunk/"), DisplayName("SVN URL")]
        public string SvnUrl
        {
            get
            {
                if (!string.IsNullOrEmpty(_svnUrl)) return _svnUrl;

                switch (Repository)
                {
                    case RepositoryType.Google:
                        _svnUrl = string.Format("http://{0}.googlecode.com/svn/trunk/", SVNProject);
                        break;

                    case RepositoryType.Assembla:
                        _svnUrl = string.Format("https://subversion.assembla.com/svn/{0}/trunk/", SVNProject);
                        break;
                }
                return _svnUrl;
            }
            set { _svnUrl = value; }
        }

        private string _changeLogUrl;
        [Category("Data"), Description("The URL of the change log without the revision number. Eg https://code.google.com/p/your-project-name/source/detail?r="), DisplayName("Change Log URL")]
        public string ChangeLogUrl
        {
            get
            {
                if (!string.IsNullOrEmpty(_changeLogUrl)) return _changeLogUrl;

                switch (Repository)
                {
                    case RepositoryType.Google:
                        _changeLogUrl = string.Format("https://code.googlecode.com/p/{0}/source/detail?r=", SVNProject);
                        break;

                    case RepositoryType.Assembla:
                        _changeLogUrl = string.Format("https://www.assembla.com/code/{0}/subversion/commit/", SVNProject);
                        break;
                }

                return _changeLogUrl;
            }
            set { _changeLogUrl = value; }
        }

        [Category("Data"), Description(@"The full path you want the files saved to. Eg: c:\HB\Combat Routines\My Routine"), DisplayName("Download Path")]
        public string DownloadPath { get; set; }

        [Category("Data"), Description("If enabled full path and file names will be shown in the log"), DisplayName("Verbose Logging"), DefaultValue(false)]
        public bool VerboseLogging { get; set; }

        [Category("Data"), Description("If enabled will not display the change log in the general HB window."), DisplayName("Silence Change Log"), DefaultValue(false)]
        public bool SilentChanngeLog { get; set; }

        [Category("Data"), Description("If enabled will not display the download messages in the general HB window."), DisplayName("Silence Downloading"), DefaultValue(false)]
        public bool SilentDownloading { get; set; }

        [Category("Data"), Description("If enabled will display a different message to indicate this update has been forced and the user does not have a choice"), DisplayName("Forced Update"), DefaultValue(false)]
        public bool ForcedUpdate { get; set; }

        [Category("Data"), Description("Friendly name of what is being downloaded, this is only used for display purposes. Eg: 'Fpsware Priest SL'")]
        public string NameOfDownload { get; set; }

        private RepositoryType _repository = RepositoryType.Google;
        [Description("The SVN repository, Google or Assembla. [default: Google]")]
        public RepositoryType Repository
        {
            get { return _repository; }
            set { _repository = value; }
        }

        public enum RepositoryType { Google, Assembla }


        private static bool _isUpdateRunning = false;
        private string _updatedFiles;
        private int _localRevision = 0;
        private int _onlineRevision;

        private readonly Regex _linkPattern = new Regex(@"<li><a href="".+"">(?<ln>.+(?:..))</a></li>", RegexOptions.CultureInvariant);

        private Regex ChangelogPattern
        {
            get
            {
                switch (Repository)
                {
                    case RepositoryType.Assembla:
                        return _changelogPatternAssembla;

                    default:
                        return _changelogPatternGoogle;
                }
            }
        }
        private readonly Regex _changelogPatternGoogle = new Regex("<h4 style=\"margin-top:0\">Log message</h4>\r?\n?<pre class=\"wrap\" style=\"margin-left:1em\">(?<log>.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?)</pre>", RegexOptions.CultureInvariant);
        private readonly Regex _changelogPatternAssembla = new Regex("<p class=\"no-margin no-border no-padding\">(?<log>.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?.+\r?\n?)</p>", RegexOptions.CultureInvariant);



        private static void Log(string message, params object[] args)
        {
            Logging.Write(LogLevel.Normal, Color.FromRgb(0, 255, 190), "{0}", string.Format(message, args));
        }

        private static void LogRed(string message, params object[] args)
        {
            Logging.Write(LogLevel.Normal, Colors.Red, "{0}", string.Format(message, args));
        }


        private void ThreadedUpdate()
        {
            _isUpdateRunning = true;

            DownloadFilesFromSvn(new WebClient(), SvnUrl);

            _isUpdateRunning = false;
            _onlineRevision.Seralize(RevisionFile);
            Log(string.Format("{0} update complete. Restart HB for the changes to take effect.", NameOfDownload));
        }

        /// <summary>
        /// Runs the updater in a new thread.
        /// </summary>
        public void AutoUpdate()
        {
            if (!_isUpdateRunning) new Thread(ThreadedAutoUpdate) { IsBackground = true }.Start();
        }

        private void ThreadedAutoUpdate()
        {
            if (IsUpdateAvailable) ThreadedUpdate();

        }

        /// <summary>
        /// Run the updater in a seperate thread
        /// </summary>
        public void DoUpdate()
        {
            if (_isUpdateRunning)
            {
                Log("An update is already in progress");
                return;
            }
            Log(string.Format("Update under way. You will be informed once this process is complete."));
            new Thread(ThreadedUpdate) { IsBackground = true }.Start();
        }


        public bool IsUpdateAvailable
        {
            get
            {
                if (_isUpdateRunning)
                {
                    Log("An update is already in progress");
                    return false;
                }

                if (String.IsNullOrEmpty(RevisionFile) || String.IsNullOrEmpty(SvnUrl) || String.IsNullOrEmpty(ChangeLogUrl) || String.IsNullOrEmpty(DownloadPath))
                {
                    throw new Exception("Required strings null or empty: SvnUrl, ChangeLogUrl, DownloadPath, RevisionFile");
                }

                if (Repository == RepositoryType.Assembla && !SvnUrl.EndsWith("/"))
                {
                    SvnUrl += "/";
                }

                if (!File.Exists(RevisionFile))
                {
                    //Log("===== revision file does not exist");
                    _localRevision.Seralize(RevisionFile);
                }
                _localRevision = Serialize.Deserialize<int>(RevisionFile);
                _onlineRevision = GetOnlineRevision();

                if (_localRevision >= _onlineRevision)
                {
                    Log(string.Format("Running revision {0} of {1}. This is the latest build. ", _localRevision, NameOfDownload));
                    return false;
                }

                if (ForcedUpdate)
                {
                    LogRed(string.Format("Build {0} of {1} will now be downloaded. This is a forced update!", _onlineRevision, NameOfDownload));
                }
                else
                {
                    Log(string.Format("Build {0} of {2} is available for download. You are running build {1}.", _onlineRevision, _localRevision, NameOfDownload));
                }

                string changelog = GetChangeLog(_onlineRevision);

                if (!SilentChanngeLog)
                {
                    if (!string.IsNullOrEmpty(changelog))
                    {
                        if (ForcedUpdate)
                        {
                            LogRed(string.Format("--------[Change Log for {0}]--------", NameOfDownload));
                            LogRed(changelog);
                        }
                        else
                        {
                            Log(string.Format("--------[Change Log for {0}]--------", NameOfDownload));
                            Log(changelog);
                        }
                        
                    }
                    else
                    {
                        Log("No change log has been provided.");
                    }
                }

                return true;
            }
        }

        public string OnlineRevision
        {
            get { return _onlineRevision.ToString(); }
        }

        public string LastUpdated { get; set; }

        public string ChangeLog
        {
            get
            {
                _onlineRevision = GetOnlineRevision();
                return GetChangeLog(_onlineRevision);
            }
        }


        private static int ActualFileCount(string url, IEnumerable<Match> matches)
        {
            return matches.Select(match => RemoveXmlEscapes(match.Groups["ln"].Value)).Select(file => url + file).Count(newUrl => newUrl[newUrl.Length - 1] != '/');
        }

        private string RelativePath(string url)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(@"..\");
            if (url.Length <= SvnUrl.Length) return sb.ToString();

            string path = url.Substring(SvnUrl.Length);
            string result = path.Replace("/", @"\");
            sb.Append(result);
            return sb.ToString();
        }

        private void DownloadFilesFromSvn(WebClient client, string url)
        {
            try
            {
                string html = client.DownloadString(url);
                MatchCollection results = _linkPattern.Matches(html);
                IEnumerable<Match> matches = from match in results.OfType<Match>() where match.Success && match.Groups["ln"].Success select match;
                if (!SilentDownloading) Log(string.Format("[{2}] Downloading {0} files in {1}", ActualFileCount(url, matches), RelativePath(url), NameOfDownload));

                //if (!matches.Any()) { return false; }

                foreach (Match match in matches)
                {
                    string file = RemoveXmlEscapes(match.Groups["ln"].Value);
                    string newUrl = url + file;
                    if (newUrl[newUrl.Length - 1] == '/') // it's a directory...
                    {
                        DownloadFilesFromSvn(client, newUrl);
                    }
                    else // its a file.
                    {
                        string filePath, dirPath;
                        if (url.Length > SvnUrl.Length)
                        {
                            string relativePath = url.Substring(SvnUrl.Length);
                            dirPath = Path.Combine(DownloadPath, relativePath);
                            filePath = Path.Combine(dirPath, file);
                        }
                        else
                        {
                            dirPath = Environment.CurrentDirectory;
                            filePath = Path.Combine(DownloadPath, file);
                        }

                        if (VerboseLogging) Log("   ..." + file);
                        _updatedFiles += (String.Format("Updated: {0}", file) + Environment.NewLine);
                        if (!Directory.Exists(dirPath)) Directory.CreateDirectory(dirPath);
                        client.DownloadFile(newUrl, filePath);
                    }
                }
            }
            catch (Exception ex) { }
        }

        private int GetOnlineRevision()
        {
            WebClient client = new WebClient();
            string html = client.DownloadString(SvnUrl);
            Regex pattern = new Regex(@" - Revision (?<rev>\d+):", RegexOptions.CultureInvariant);
            Match match = pattern.Match(html);
            if (match.Success && match.Groups["rev"].Success)
            {               
                return int.Parse(match.Groups["rev"].Value);
            }

            Log("OH SHIT! Unable to retreive online revision");
            throw new Exception("Unable to retreive revision! The sky is falling!");
        }

        private static string RemoveXmlEscapes(string xml)
        {
            return xml.Replace("&amp;", "&").Replace("&lt;", "<").Replace("&gt;", ">").Replace("&quot;", "\"").Replace("&apos;", "'").Replace("<br />", "\n");
        }

        private string GetChangeLog(int revision)
        {
            var client = new WebClient();
            string html = client.DownloadString(ChangeLogUrl + revision);
            Match match = ChangelogPattern.Match(html);
            if (match.Success && match.Groups["log"].Success)
            {
                string changelog = RemoveXmlEscapes(match.Groups["log"].Value);

                if (Repository == RepositoryType.Assembla)
                {
                    Regex lastUpdated = new Regex(@"(?<=span>\().*?(?=[\)])", RegexOptions.CultureInvariant);
                    match = lastUpdated.Match(html);
                    if (match.Success)
                    {
                        LastUpdated = match.Value;
                    }
                }

                return changelog;
            }
            return null;
        }
    }

}
