﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ComponentModel;
using System.Reflection;
using System.Windows.Forms;
using Fpsware.Library3.UI;
using Styx.CommonBot;
using Styx.Helpers;

namespace Fpsware.Library3.Helpers
{
    public class UISettingsPopulater 
    {
        //public UISettingsPopulater() { InitializeComponent(); }
        //public UISettingsPopulater(IContainer container) { container.Add(this); InitializeComponent(); }

        //public Settings SettingInstance { get; set; }
        public Settings SettingInstance { get; set; }
        public Type SettingType { get; set; }
        public Control HostControl { get; set; }
        public bool IsLoading { get; set; }

        // Used for our delayed save
        readonly Timer _savetimer = new Timer();

        /// <summary>
        /// Returns a collection of controls matching the type passed. Will work recursively.
        /// </summary>
        /// <param name="hostControl">The starting point, the control that contains all your controls. Eg a GroupBox or a Form</param>
        /// <param name="type">The type of controls you want to find. Eg typeof(Textbox)</param>
        /// <returns>Returns the collection of controls of the type requested. Eg a collection of all Textboxes on the form</returns>
        public IEnumerable<Control> GetAllControlsOfType(Control hostControl, Type type)
        {
            IEnumerable<Control> controls = hostControl.Controls.Cast<Control>();
            return controls.SelectMany(ctrl => GetAllControlsOfType(ctrl, type)).Concat(controls).Where(c => c.GetType() == type);
        }

        public void Go()
        {
            // Setup our timed setting save.
            // Some controls such as Numeric Up Down and Progress Bar trigger the value changed event for every small increment. 
            // We want to delay the save until at least 1 second has elapsed since any change. 
            _savetimer.Interval = 1000;
            _savetimer.Tick += (sender, args) =>
            {
                //Log.Info("Setting saved!");
                SettingInstance.Save();
                _savetimer.Enabled = false;
                _savetimer.Stop();
            };

            _savetimer.Enabled = false;
            _savetimer.Stop();


            // Get all properties from the Setting class
            PropertyInfo[] settingsProperties = SettingType.GetProperties();

            // Get a list of all our appropriate controls in the HostControl (typically this will be the UI form)
            IEnumerable<Control> dropdownControls = GetAllControlsOfType(HostControl, typeof(FPSDropdown));
            IEnumerable<Control> numControls = GetAllControlsOfType(HostControl, typeof(FPSNumericUpdown));
            IEnumerable<Control> progressbarControls = GetAllControlsOfType(HostControl, typeof(FPSProgressBar));
            IEnumerable<Control> textboxControls = GetAllControlsOfType(HostControl, typeof(FPSTextBox));
            //IEnumerable<Control> toggleControls = GetAllControlsOfType(HostControl, typeof(FPSToggleButton));


            // Enumerate every setting property
            // Eg... this is a setting property Setting.Instance.ImASettingProperty
            foreach (PropertyInfo setting in settingsProperties)
            {
                PropertyInfo pInfo = SettingType.GetProperty(setting.Name);

                // This is the current value of the property.
                object propValue = pInfo.GetValue(SettingInstance, null);

                // Enumerate all dropdown controls, where the Dropdown.SettingName == Setting.Name
                // This should only return ONE control with this Setting.name
                foreach (FPSDropdown dropdown in dropdownControls.Select(control => control as FPSDropdown).Where(dropdown => dropdown.SettingName == setting.Name))
                {

                    // If this control has a spell associated with it check if Spellmanager knows the spell
                    // If not, grey out the control to indicate its not useable
                    if (!string.IsNullOrEmpty(dropdown.SpellName)) dropdown.IsSpellValid = SpellManager.HasSpell(dropdown.SpellName);

                    // Enumerate all attributes associated with this property and pull out what we need
                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    
                    Attributes.DropdownItems dropdownItems = (Attributes.DropdownItems)attributes[typeof(Attributes.DropdownItems)];
                    if (dropdownItems != null && dropdownItems.Values.Any())
                    {
                        List<string> ddItems = dropdown.DropDownItems.Concat(dropdownItems.Values).Distinct().ToList();
                        //dropdown.Dropdown.Items.Clear();
                        dropdown.ClearDropdownOnly();
                        //dropdown.Dropdown.Items.AddRange(ddItems.ToArray());
                        dropdown.AddRange(ddItems.ToArray());
                    }

                    // Assign the Description attribute to the tooltip text 
                    if (da != null && !string.IsNullOrEmpty(da.Description))
                    {
                        dropdown.TooltipText = da.Description;
                        dropdown.TooltipTitle = !string.IsNullOrEmpty(dropdown.SpellName) ? dropdown.SpellName : dropdown.Label;
                    }
                    // A second check if this property requires a spell to be in the spellmanager (but this one is based on the attribute in the settings property)
                    // If this control has a spell associated with it if Spellmanager knows the spell
                    // If not, grey out the control to indicate its not useable
                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) dropdown.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    // Add the value changed event
                    dropdown.ValueChanged += FPSDropdown_ValueChangedEvent;
                    

                    try
                    {
                        // Set the current value
                        dropdown.Value = propValue.ToString();
                    }
                    catch (Exception e) {  }

                }

                // Progress Bars
                foreach (FPSProgressBar bar in progressbarControls.Select(control => control as FPSProgressBar).Where(bar => bar.SettingName == setting.Name))
                {
                    bar.ValueChanged += FPSProgressBar_ValueChangedEvent;
                    bar.Value = (int)propValue;

                    if (!string.IsNullOrEmpty(bar.SpellName)) bar.IsSpellValid = SpellManager.HasSpell(bar.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) bar.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    bar.TooltipText = da.Description;
                    bar.TooltipTitle = !string.IsNullOrEmpty(bar.SpellName) ? bar.SpellName : bar.Label;
                }

                // Numeric Up Down
                foreach (FPSNumericUpdown num in numControls.Select(control => control as FPSNumericUpdown).Where(bar => bar.SettingName == setting.Name))
                {
                    num.ValueChanged += FPSNumericUpDown_ValueChangedEvent;
                    num.Value = (int)propValue;

                    if (!string.IsNullOrEmpty(num.SpellName)) num.IsSpellValid = SpellManager.HasSpell(num.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) num.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    num.TooltipText = da.Description;
                    num.TooltipTitle = !string.IsNullOrEmpty(num.SpellName) ? num.SpellName : num.Label;
                }

                // Textbox
                foreach (FPSTextBox textBox in textboxControls.Select(control => control as FPSTextBox).Where(bar => bar.SettingName == setting.Name))
                {
                    textBox.ValueChanged += FPSTextBox_ValueChangedEvent;
                    textBox.Value = propValue.ToString();

                    if (!string.IsNullOrEmpty(textBox.SpellName)) textBox.IsSpellValid = SpellManager.HasSpell(textBox.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) textBox.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    textBox.TooltipText = da.Description;
                    textBox.TooltipTitle = !string.IsNullOrEmpty(textBox.SpellName) ? textBox.SpellName : textBox.Label;
                }


            }

        }


        // These are the events associated with all of our controls

        private void FPSDropdown_ValueChangedEvent(object sender, DropdownValueChangedEventArgs e)
        {
            // If we're still loading our controls then don't fire this event if the value changes
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Save the changes immediately
                SettingInstance.Save();
            }

        }

        private void FPSProgressBar_ValueChangedEvent(object sender, ProgressBarValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }

        private void FPSNumericUpDown_ValueChangedEvent(object sender, NumbericUpDownValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }

        private void FPSTextBox_ValueChangedEvent(object sender, TextBoxValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }

        /*
        private void FPSToggle_ValueChangedEvent(object sender, ToggleValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

            }
        }
         */
    }


    static public class Attributes
    {

        public enum ControlTypeEnum
        {
            Dropdown,
            ProgressBar,
            NumericUpDown,
            Textbox
            //Toggle
        }

        /// <summary>
        /// Type of control for the UI popup menu. Options are: Dropdown, ProgressBar, NumericUpDown, Textbox. If this attribute is omitted the default of Dropdown will be assumed.
        /// </summary>
        public class ControlType : Attribute
        {
            public ControlTypeEnum Name { get; set; }

            public ControlType(ControlTypeEnum controlType)
            {
                Name = controlType;
            }

        }

        public enum PowerTypeEnum
        {
            Health,
            Mana,
            Energy,
            Rage,
        }

        /// <summary>
        /// The power type this control will represent. This will determine the colour of the progress bar
        /// </summary>
        public class PowerType : Attribute
        {
            public PowerTypeEnum Name { get; set; }

            public PowerType(PowerTypeEnum powerType)
            {
                Name = powerType;
            }

        }







        /// <summary>
        /// This spell is required in the SpellManager for this control to be enabled.
        /// </summary>
        public class SpellRequired : Attribute
        {
            public SpellRequired(string spellName)
            {
                Name = spellName;
            }

            public string Name { get; set; }
        }

        public class DropdownItems : Attribute
        {
            public string[] Values { get; set; }

            public DropdownItems(params string[] values)
            {
                Values = values;
            }
        }

        /// <summary>
        /// Used for both Numberic Up Down and Progress Bar controls
        /// </summary>
        public class MinMaxRange : Attribute
        {
            public MinMaxRange(int minimumRange, int maximumRange)
            {
                MinimumRange = minimumRange;
                MaximumRange = maximumRange;
            }

            public int MinimumRange { get; set; }
            public int MaximumRange { get; set; }
        }

    }
}
