﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using Fpsware.Library3.UI.Cyotek.Design;

namespace Fpsware.Library3.UI.Cyotek
{
  [ToolboxItem(true)]
  [Designer(typeof(TabListDesigner))]
  [DefaultProperty("TabListPages"), DefaultEvent("SelectedIndexChanged")]
  public partial class TabList : Control
  {
    #region  Private Member Declarations

    private Rectangle _displayRectangle;
    private Size _headerSize;
    private int _hoverIndex;
    private TabListPage[] _pages;
    private ITabListPageRenderer _renderer;
    private int _selectedIndex;

      #endregion  Private Member Declarations

    #region  Static Constructors

    static TabList()
    {
      TabListPageRenderer.DefaultRenderer = new DefaultTabListPageRenderer();
    }

    #endregion  Static Constructors

    #region  Public Constructors

    public TabList()
    {
      // ReSharper disable DoNotCallOverridableMethodsInConstructor
      TabListPages = new TabListPageCollection(this);
      SelectedIndex = -1;
      HoverIndex = -1;
      HeaderSize = new Size(150, 25);
      Size = new Size(200, 200); // the default size is tiny!
      Padding = new Padding(3);
      DoubleBuffered = true;
      ShowTabList = true;
      AllowTabSelection = true;
      SetStyles();
      // ReSharper restore DoNotCallOverridableMethodsInConstructor
    }

    protected virtual void SetStyles()
    {
      SetStyle(ControlStyles.AllPaintingInWmPaint
                        | ControlStyles.ResizeRedraw
                        | ControlStyles.OptimizedDoubleBuffer
                        | ControlStyles.UserMouse
                        , true);
      SetStyle(ControlStyles.Selectable, AllowTabSelection);
    }

    #endregion  Public Constructors

    #region  Events

    [Category("Property Changed")]
    public event EventHandler HeaderSizeChanged;

    [Category("Property Changed")]
    public event EventHandler RendererChanged;

    [Category("Property Changed")]
    public event EventHandler SelectedIndexChanged;

    #endregion  Events

    #region  Overriden Properties

    public override Rectangle DisplayRectangle
    {
      get
      {
        if (_displayRectangle.IsEmpty)
        {
          int leftMargin;

            // to avoid calculating every time, calculate it only when it's need it and cache it like any normal property
          if (ShowTabList)
            leftMargin = HeaderSize.Width + Padding.Left + 1;
          else
            leftMargin = Padding.Left + 1;
          int topMargin = Padding.Top + 1;
          int rightMargin = Padding.Right + 1;
          int bottomMargin = Padding.Bottom + 1;

          _displayRectangle = new Rectangle(leftMargin, topMargin, ClientSize.Width - (leftMargin + rightMargin), ClientSize.Height - (topMargin + bottomMargin));
        }

        return _displayRectangle;
      }
    }

    [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public override string Text
    {
      get { return base.Text; }
      set { base.Text = value; }
    }

    #endregion  Overriden Properties

    #region  Protected Overridden Methods

    protected override ControlCollection CreateControlsInstance()
    {
      return new TabListControlCollection(this);
    }

    protected override bool IsInputKey(Keys keyData)
    {
      bool result;

      switch (keyData)
      {
        case Keys.Up:
        case Keys.Down:
          result = true;
          break;
        default:
          result = base.IsInputKey(keyData);
          break;
      }

      return result;
    }

    protected override void OnGotFocus(EventArgs e)
    {
      base.OnGotFocus(e);

      Invalidate();
    }

    protected override void OnKeyDown(KeyEventArgs e)
    {
      base.OnKeyDown(e);

      if (ShowTabList)
      {
        // allow keyboard navigation, if any tabs are present
        if (TabListPageCount != 0)
        {
          switch (e.KeyCode)
          {
            case Keys.Down:
              CycleSelectedTab(1);
              break;
            case Keys.Up:
              CycleSelectedTab(-1);
              break;
            case Keys.PageDown:
              CycleSelectedTab(3);
              break;
            case Keys.PageUp:
              CycleSelectedTab(-3);
              break;
            case Keys.Home:
              SelectedIndex = 0;
              break;
            case Keys.End:
              SelectedIndex = TabListPageCount - 1;
              break;
          }
        }
      }
    }

    protected override void OnLostFocus(EventArgs e)
    {
      base.OnLostFocus(e);

      Invalidate();
    }

    protected override void OnMouseClick(MouseEventArgs e)
    {
      base.OnMouseClick(e);

      if (e.Button == MouseButtons.Left && (DesignMode | AllowTabSelection))
      {
          int index = GetItemAtPoint(e.Location);
          if (index != -1)
          SelectedIndex = index;
      }
    }

    protected override void OnMouseLeave(EventArgs e)
    {
      base.OnMouseLeave(e);

      HoverIndex = -1;
    }

    protected override void OnMouseMove(MouseEventArgs e)
    {
      base.OnMouseMove(e);

      if (AllowTabSelection)
        HoverIndex = GetItemAtPoint(e.Location);
    }

    protected override void OnPaddingChanged(EventArgs e)
    {
      base.OnPaddingChanged(e);

      ResetSelectedPage();
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);

      if (ShowTabList)
      {
          ITabListPageRenderer renderer = GetRenderer();

          for (int i = 0; i < TabListPageCount; i++)
        {
          TabListPageState state;

          if (i == SelectedIndex)
          {
            state = TabListPageState.Selected;
            if (Focused)
              state |= TabListPageState.Focused;
          }
          else if (i == HoverIndex)
            state = TabListPageState.Hot;
          else
            state = TabListPageState.Normal;

          renderer.RenderHeader(e.Graphics, TabListPages[i], state);
        }
      }
    }

    protected override void OnResize(EventArgs e)
    {
      base.OnResize(e);

      ResetSelectedPage();
    }

    #endregion  Protected Overridden Methods

    #region  Public Methods

    public virtual TabListPage HitTest(Point point)
    {
        int index = GetItemAtPoint(point);

        return index != -1 ? TabListPages[index] : null;
    }

      #endregion  Public Methods

    #region  Internal Methods

    internal int AddPage(TabListPage page)
    {
        int index = InsertPage(TabListPageCount, page);

      if (SelectedIndex == -1)
        SelectedIndex = index;

      return index;
    }

    internal void ClearAllPages()
    {
      Controls.Clear();
      _pages = null;
      TabListPageCount = 0;
    }

    internal TabListPage[] GetTabListPages()
    {
        TabListPage[] copy = new TabListPage[TabListPageCount];

      if (TabListPageCount > 0)
        Array.Copy(_pages, copy, TabListPageCount);

      return copy;
    }

    internal int InsertPage(int index, TabListPage page)
    {
      if (_pages == null)
        _pages = new TabListPage[1];
      else if (_pages.Length == TabListPageCount)
      {
        // no room left, so resize the array

          TabListPage[] copy = new TabListPage[_pages.Length + 1];
        Array.Copy(_pages, copy, _pages.Length);
        _pages = copy;
      }

      // if this is an insert rather than append, move the array around
      if (index < TabListPageCount)
        Array.Copy(_pages, index, _pages, index + 1, _pages.Length - index);

      // update the array and page count
      _pages[index] = page;
      TabListPageCount++;
      UpdatePages();
      UpdateSelectedPage();

      // finally trigger a redraw of the control
      Invalidate();

      return index;
    }

      public void RemovePageAt(int index)
    {
        if ((index < 0) || (index >= TabListPageCount))
        throw new ArgumentOutOfRangeException("index", "");

      TabListPageCount--;

      if (index < TabListPageCount)
        Array.Copy(_pages, index + 1, _pages, index, TabListPageCount - index);

      _pages[TabListPageCount] = null;

      int selectedIndex = SelectedIndex;
      if (TabListPageCount == 0)
        SelectedIndex = -1;
      else if (index == selectedIndex || selectedIndex >= TabListPageCount)
        SelectedIndex = 0;
      UpdatePages();
      UpdateSelectedPage();

      Invalidate();
    }

    internal void UpdatePage(TabListPage page)
    {
      Invalidate();
    }

    #endregion  Internal Methods

    #region  Public Properties

      /*
      public Color BackColorHeader
      {
          get { return _backColorHeader; }
          set { _backColorHeader = value; }
      }
       */


      [Category("Appearance"), DefaultValue(typeof(Size), "150, 25")]
    public virtual Size HeaderSize
    {
      get { return _headerSize; }
      set
      {
        if (HeaderSize != value)
        {
          _headerSize = value;

          OnHeaderSizeChanged(EventArgs.Empty);
        }
      }
    }

    [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public virtual int HoverIndex
    {
      get { return _hoverIndex; }
      set
      {
        if (_hoverIndex != value)
        {
          _hoverIndex = value;
          Invalidate();
        }
      }
    }

    [DefaultValue(typeof(Padding), "3, 3, 3, 3")]
    public new Padding Padding
    {
      get { return base.Padding; }
      set { base.Padding = value; }
    }

    [Category("Appearance"), DefaultValue(null)]
    public virtual ITabListPageRenderer Renderer
    {
      get { return _renderer; }
      set
      {
        if (Renderer != value)
        {
          _renderer = value;

          OnRendererChanged(EventArgs.Empty);
        }
      }
    }

    [Browsable(false), DefaultValue(-1), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public virtual int SelectedIndex
    {
      get { return _selectedIndex; }
      set
      {
        if (SelectedIndex != value)
        {
          _selectedIndex = value;

          OnSelectedIndexChanged(EventArgs.Empty);
        }
      }
    }

    [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public virtual TabListPage SelectedPage
    {
      get { return SelectedIndex != -1 ? TabListPages[SelectedIndex] : null; }
      set { SelectedIndex = TabListPages.IndexOf(value); }
    }

      [Browsable(false)]
      public virtual int TabListPageCount { get; protected set; }

      [Category("Behavior"), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
      public virtual TabListPageCollection TabListPages { get; protected set; }

      #endregion  Public Properties

    #region  Protected Methods

    protected virtual void CycleSelectedTab(int increment)
    {
      if (TabListPageCount != 0)
      {
          int index = SelectedIndex + increment;
        if (index < 0 && increment == -1 || index >= TabListPageCount && increment != 1)
          index = TabListPageCount - 1;
        else if (index < 0 && increment != -1 || index >= TabListPageCount && increment == 1)
          index = 0;

        SelectedIndex = index;
      }
    }

    protected virtual int GetItemAtPoint(Point point)
    {
        int result = -1;

      for (int i = 0; i < TabListPageCount; i++)
      {
        if (TabListPages[i].HeaderBounds.Contains(point))
        {
          result = i;
          break;
        }
      }

      return result;
    }

    protected virtual ITabListPageRenderer GetRenderer()
    {
      return Renderer ?? TabListPageRenderer.DefaultRenderer ?? new DefaultTabListPageRenderer();
    }

    protected virtual void OnHeaderSizeChanged(EventArgs e)
    {
        ResetSelectedPage();

      EventHandler handler = HeaderSizeChanged;

      if (handler != null)
        handler(this, e);
    }

    protected virtual void OnRendererChanged(EventArgs e)
    {
        Invalidate();

      EventHandler handler = RendererChanged;

      if (handler != null)
        handler(this, e);
    }

    protected virtual void OnSelectedIndexChanged(EventArgs e)
    {
        UpdateSelectedPage();

      EventHandler handler = SelectedIndexChanged;

      if (handler != null)
        handler(this, e);
    }

    protected virtual void ResetSelectedPage()
    {
      _displayRectangle = Rectangle.Empty; // force the display rectangle to be recalculated
      UpdateSelectedPage();
    }

    protected virtual void UpdatePages()
    {
        int y = Padding.Top;
      int x = Padding.Left;

      foreach (TabListPage page in TabListPages)
      {
        page.HeaderBounds = new Rectangle(new Point(x, y), HeaderSize);
        y += HeaderSize.Height;
      }
    }

    protected virtual void UpdateSelectedPage()
    {
      if (SelectedIndex != -1)
      {
        SelectedPage.Bounds = DisplayRectangle;

        for (int i = 0; i < TabListPageCount; i++)
          _pages[i].Visible = (i == SelectedIndex);
      }

      Invalidate();
    }

    #endregion  Protected Methods

    private bool _showTabList;

    [Category("Appearance"), DefaultValue(true)]
    public virtual bool ShowTabList
    {
      get { return _showTabList; }
      set
      {
        if (ShowTabList != value)
        {
          _showTabList = value;

          OnShowTabListChanged(EventArgs.Empty);
        }
      }
    }

    /// <summary>
    /// Occurs when the ShowTabList property value changes
    /// </summary>
    [Category("Property Changed")]
    public event EventHandler ShowTabListChanged;

    /// <summary>
    /// Raises the <see cref="ShowTabListChanged" /> event.
    /// </summary>
    /// <param name="e">The <see cref="EventArgs" /> instance containing the event data.</param>
    protected virtual void OnShowTabListChanged(EventArgs e)
    {
        ResetSelectedPage();

      EventHandler handler = ShowTabListChanged;

      if (handler != null)
        handler(this, e);
    }

    private bool _allowTabSelection;
      private Color _backColorHeader;

      [Category("Behavior"), DefaultValue(true)]
    public virtual bool AllowTabSelection
    {
      get { return _allowTabSelection; }
      set
      {
        if (AllowTabSelection != value)
        {
          _allowTabSelection = value;

          OnAllowTabSelectionChanged(EventArgs.Empty);
        }
      }
    }

    /// <summary>
    /// Occurs when the AllowTabSelection property value changes
    /// </summary>
    [Category("Property Changed")]
    public event EventHandler AllowTabSelectionChanged;

    /// <summary>
    /// Raises the <see cref="AllowTabSelectionChanged" /> event.
    /// </summary>
    /// <param name="e">The <see cref="EventArgs" /> instance containing the event data.</param>
    protected virtual void OnAllowTabSelectionChanged(EventArgs e)
    {
        SetStyles();

      EventHandler handler = AllowTabSelectionChanged;

      if (handler != null)
        handler(this, e);
    }


  }
}
