﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace Fpsware.Library3.UI
{
    public partial class FPSProgressBar : UserControl
    {
        private string _toolTipText;
        private int _oldValue;
        private string _settingName;
        private bool _additionalsVisible;
        private Color _barColor = Color.LimeGreen;

        // Event stuff when value is changed
        public delegate void LineEventHandler(object sender, ProgressBarValueChangedEventArgs e);
        public event LineEventHandler ValueChanged;
        protected virtual void OnValueChanged(ProgressBarValueChangedEventArgs e) { try { ValueChanged(this, e); } catch (Exception exception) { } }

        public delegate void AdditionalsEventHandler(object sender, AddingNewEventArgs e);
        public event AdditionalsEventHandler AdditionalsButtonPressed;
        protected virtual void OnAdditionalsButtonPressed(AddingNewEventArgs e) { try { AdditionalsButtonPressed(this, e); } catch { } }

        public FPSProgressBar(string label, string settingName, int value)
        {
            Label = label;
            SettingName = settingName;
            Value = value;
        }

        public FPSProgressBar(string label, string settingName, int value, string tooltipTitle, string tooltipText)
        {
            Label = label;
            SettingName = settingName;
            Value = value;
            TooltipTitle = tooltipTitle;
            TooltipText = tooltipText;
        }

        public FPSProgressBar() { InitializeComponent(); }

        [Category("_Fpsware")]
        [Description("Gets or sets the spell associated with this control"), DefaultValue("")]
        public string SpellName { get; set; }

        [Category("_Fpsware")]
        [Description("Sets if the the spell exists in the spellmanager"), DefaultValue(true)]
        public bool IsSpellValid
        {
            set { lblText.ForeColor = value ? Color.Black : Color.Silver; }
        }

        private void ResizeDropdown()
        {
            pbValue.Left = lblText.Width + 9;

            if (!AdditionalsAvailable)
            {
                pbValue.Width = Width - pbValue.Location.X - 3;
            }
            else
            {
                pbValue.Width = Width - ((pbValue.Location.X - 3) + (btnAdditionals.Width + 10));
            }
        }

        [Category("_Fpsware")]
        [Description("Show or hide the 'additionals' button")]
        [DefaultValue(false)]
        public bool AdditionalsAvailable
        {
            get { return _additionalsVisible; }
            set
            {
                _additionalsVisible = value;

                if (_additionalsVisible)
                {
                    btnAdditionals.Visible = true;
                    ResizeDropdown();
                }
                else
                {
                    btnAdditionals.Visible = false;
                    pbValue.Width = ((Width - pbValue.Left - 3));
                }
            }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the label of this control")]
        [DefaultValue("Description")]
        public string Label
        {
            get { return lblText.Text; }
            set { lblText.Text = value; }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the value of this control")]
        [DefaultValue(50)]
        public int Value
        {
            get { return (int)pbValue.Value; }
            set
            {
                pbValue.Value = value;
            }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the maximum value allowed")]
        [DefaultValue(100)]
        public int MaxValue
        {
            get { return (int)pbValue.Maximum; }
            set { pbValue.Maximum = value; }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the minimum value allowed")]
        [DefaultValue(0)]
        public int MinValue
        {
            get { return (int)pbValue.Minimum; }
            set { pbValue.Minimum = value; }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the associated setting")]
        public string SettingName
        {
            get { return _settingName ?? "No setting or null"; }
            set { _settingName = value; }
        }

        public Color BarColor
        {
            get { return _barColor; }
            set
            {
                _barColor = value;
                pbValue.ForeColor = value;
            }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the tooltip text")]
        [DefaultValue("Fpsware CC")]
        public string TooltipText
        {
            get { return _toolTipText; }
            set
            {
                if (value == null) return;

                _toolTipText = value;
                toolTip1.SetToolTip(pbValue, _toolTipText);
            }
        }

        [Category("_Fpsware")]
        [Description("Gets or sets the title used in the tooltip")]
        public string TooltipTitle
        {
            get { return toolTip1.ToolTipTitle; }
            set
            {
                if (value == null) return;
                toolTip1.ToolTipTitle = value;
            }
        }

        [Category("_Fpsware")]
        [Description("Enables or disables user changes")]
        public new bool Enabled
        {
            get { return pbValue.Enabled; }
            set
            {
                pbValue.Enabled = value;
                lblText.Enabled = value;
            }
        }

        [Category("_Fpsware")]
        [Description("Sets the width of the label, thus reducing the size of the progress bar")]
        [DefaultValue(110)]
        public int LabelSize
        {
            get { return lblText.Width; }
            set
            {
                lblText.Width = value;
                pbValue.Left = lblText.Width + 9;
                pbValue.Width = (Width - pbValue.Left - 3);
                ResizeDropdown();
            }
        }

        private void btnAdditionals_Click(object sender, EventArgs e)
        {
            OnAdditionalsButtonPressed(new AddingNewEventArgs(SettingName));
        }

        private void fpsProgressBarExt1_ValueChanged_1(object sender, FPSProgressBarExt.ProgressBarEventArgs e)
        {
            Value = (int)pbValue.Value;

            // Trigger the 'value changed' event
            OnValueChanged(new ProgressBarValueChangedEventArgs(Value, SettingName));
        }

    }

    public class ProgressBarValueChangedEventArgs : EventArgs
    {
        public ProgressBarValueChangedEventArgs(int value, string settingName)
        {
            SettingName = settingName;
            Value = value;
        }

        public int Value { get; private set; }
        public string SettingName { get; private set; }
    }
}
