﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Reflection;
using System.Windows.Forms;
using Fpsware.Library3.Extensions;
using Fpsware.Library3.Helpers;
using Fpsware.Library3.UI.Popup;
using Fpsware.Library3.Utility;
using Styx.CommonBot;
using Styx.Helpers;

namespace Fpsware.Library3.UI
{
    public partial class UIAdditionals : UserControl
    {
        public UIAdditionals() { InitializeComponent();  }

        public Settings SettingInstance { get; set; }
        public Type SettingType { get; set; }
        public bool IsLoading { get; set; }
        public PopupComboBox PopupComboBoxParent { get; set; }

        private void UIAdditionals_Load(object sender, EventArgs e)
        {
            try { PopulateUIControls(); } catch (Exception exception) {  }

            OrganiseAllControls();

        }

        /// <summary>
        /// Return the number of controlls added to the popup. 
        /// </summary>
        public int CountOfControls
        {
            get
            {
                return panel1.Controls.Count;
            }
        }

        #region Standard UI population stuff


        readonly Timer _savetimer = new Timer();


        /// <summary>
        /// Returns a collection of controls matching the type passed. Will work recursively.
        /// </summary>
        /// <param name="hostControl">The starting point, the control that contains all your controls. Eg a GroupBox or a Form</param>
        /// <param name="type">The type of controls you want to find. Eg typeof(Textbox)</param>
        /// <returns>Returns the collection of controls of the type requested. Eg a collection of all Textboxes on the form</returns>
        public IEnumerable<Control> GetAllControlsOfType(Control hostControl, Type type)
        {
            IEnumerable<Control> controls = hostControl.Controls.Cast<Control>();
            return controls.SelectMany(ctrl => GetAllControlsOfType(ctrl, type)).Concat(controls).Where(c => c.GetType() == type);
        }

        public void PopulateUIControls()
        {
            // Setup our timed setting save.
            // Some controls such as Numeric Up Down and Progress Bar trigger the value changed event for every small increment. 
            // We want to delay the save until at least 1.5 seconds has elapsed since any change. 
            _savetimer.Interval = 1000;
            _savetimer.Tick += (sender, args) =>
            {
                //Log.Info("Setting saved!");
                SettingInstance.Save();
                _savetimer.Enabled = false;
                _savetimer.Stop();


            };

            _savetimer.Enabled = false;
            _savetimer.Stop();


            // Get all properties from the Setting class
            PropertyInfo[] settingsProperties = SettingType.GetProperties();

            // Get a list of all our appropriate controls on the firm
            IEnumerable<Control> dropdownControls = GetAllControlsOfType(panel1, typeof(FPSDropdown));
            IEnumerable<Control> numControls = GetAllControlsOfType(panel1, typeof(FPSNumericUpdown));
            IEnumerable<Control> progressbarControls = GetAllControlsOfType(panel1, typeof(FPSProgressBar));
            IEnumerable<Control> textboxControls = GetAllControlsOfType(panel1, typeof(FPSTextBox));


            // Enumerate every setting property
            foreach (PropertyInfo setting in settingsProperties)
            {
                PropertyInfo pInfo = SettingType.GetProperty(setting.Name);

                // This is the current value of the property.
                object propValue = pInfo.GetValue(SettingInstance, null);

                // Enumerate all dropdown controls, where the Dropdown.SettingName == Setting.Name
                // This should only return ONE control with this Setting.name
                foreach (FPSDropdown dropdown in dropdownControls.Select(control => control as FPSDropdown).Where(dropdown => dropdown.SettingName == setting.Name))
                {
                    // Add the value changed event
                    dropdown.ValueChanged += FPSDropdown_ValueChangedEvent;

                    // Set the current value
                    dropdown.Value = propValue.ToString();

                    // If this control has a spell associated with it check if Spellmanager knows the spell
                    // If not, grey out the control to indicate its not useable
                    if (!string.IsNullOrEmpty(dropdown.SpellName)) dropdown.IsSpellValid = SpellManager.HasSpell(dropdown.SpellName);

                    // Enumerate all attributes associated with this property and pull out what we need
                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    // Assign the Description attribute to the tooltip text 
                    if (da != null && !string.IsNullOrEmpty(da.Description))
                    {
                        dropdown.TooltipText = da.Description;
                        dropdown.TooltipTitle = !string.IsNullOrEmpty(dropdown.SpellName) ? dropdown.SpellName : dropdown.Label;
                    }

                    // A second check if this property requires a spell to be in the spellmanager (but this one is based on the attribute in the settings property)
                    // If this control has a spell associated with it if Spellmanager knows the spell
                    // If not, grey out the control to indicate its not useable
                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) dropdown.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);
                }


                // Progress Bars
                foreach (FPSProgressBar bar in progressbarControls.Select(control => control as FPSProgressBar).Where(bar => bar.SettingName == setting.Name))
                {
                    bar.ValueChanged += FPSProgressBar_ValueChangedEvent;
                    bar.Value = (int)propValue;

                    if (!string.IsNullOrEmpty(bar.SpellName)) bar.IsSpellValid = SpellManager.HasSpell(bar.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) bar.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    bar.TooltipText = da.Description;
                    bar.TooltipTitle = !string.IsNullOrEmpty(bar.SpellName) ? bar.SpellName : bar.Label;
                }

                // Numeric Up Down
                foreach (FPSNumericUpdown num in numControls.Select(control => control as FPSNumericUpdown).Where(bar => bar.SettingName == setting.Name))
                {
                    num.ValueChanged += FPSNumericUpDown_ValueChangedEvent;
                    num.Value = (int)propValue;

                    if (!string.IsNullOrEmpty(num.SpellName)) num.IsSpellValid = SpellManager.HasSpell(num.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) num.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    num.TooltipText = da.Description;
                    num.TooltipTitle = !string.IsNullOrEmpty(num.SpellName) ? num.SpellName : num.Label;
                }

                // Textbox
                foreach (FPSTextBox textBox in textboxControls.Select(control => control as FPSTextBox).Where(bar => bar.SettingName == setting.Name))
                {
                    textBox.ValueChanged += FPSTextBox_ValueChangedEvent;
                    textBox.Value = propValue.ToString();

                    if (!string.IsNullOrEmpty(textBox.SpellName)) textBox.IsSpellValid = SpellManager.HasSpell(textBox.SpellName);

                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    DescriptionAttribute da = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
                    Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

                    if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name)) textBox.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);

                    if (da == null || string.IsNullOrEmpty(da.Description)) continue;
                    textBox.TooltipText = da.Description;
                    textBox.TooltipTitle = !string.IsNullOrEmpty(textBox.SpellName) ? textBox.SpellName : textBox.Label;
                }


            }

        }


        // These are the events associated with all of our controls

        private void FPSDropdown_ValueChangedEvent(object sender, DropdownValueChangedEventArgs e)
        {
            // If we're still loading our controls then don't fire this event if the value changes
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Save the changes immediately
                SettingInstance.Save();
            }

        }

        private void FPSProgressBar_ValueChangedEvent(object sender, ProgressBarValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }

        private void FPSNumericUpDown_ValueChangedEvent(object sender, NumbericUpDownValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }

        private void FPSTextBox_ValueChangedEvent(object sender, TextBoxValueChangedEventArgs e)
        {
            if (IsLoading) return;

            // Get the first property in the Setting class with the name that matches the changed setting. Eg Setting.Instance.MightyBash == e.SettingName ("MightyBash")
            PropertyInfo settingProperty = SettingType.GetProperties().FirstOrDefault(p => String.Equals(p.Name, e.SettingName, StringComparison.CurrentCultureIgnoreCase));
            if (settingProperty != null)
            {
                settingProperty.SetValue(SettingInstance, e.Value, null);

                // Enable our delayed save
                _savetimer.Stop();
                _savetimer.Enabled = true;
                _savetimer.Start();

            }
        }


        #endregion

        private void button1_Click(object sender, EventArgs e)
        {
            if (PopupComboBoxParent == null) Dispose(); else { PopupComboBoxParent.HideDropDown(); }
        }

        public void AddCategory(string categoryName)
        {
            // Get all properties from the Setting class
            PropertyInfo[] settingsProperties = SettingType.GetProperties();
            
            // Enumerate every setting property and find those with a Category Attribute defined
            // If it has a category attribute and the name = categoryName then add the new control
            foreach (PropertyInfo setting in settingsProperties)
            {
                try
                {
                    AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[setting.Name].Attributes;
                    CategoryAttribute category = (CategoryAttribute)attributes[typeof(CategoryAttribute)];

                    // No category attribute? Then keep going
                    if (category == null || category.Category != categoryName) continue;

                    // Category attribute found. 
                    // Lets see what type of control this is going to be
                    Attributes.ControlType controlType = (Attributes.ControlType)attributes[typeof(Attributes.ControlType)];

                    //if (controlType != null && !string.IsNullOrEmpty(controlType.Name))
                    if (controlType != null)
                    {
                        switch (controlType.Name)
                        {
                            case Attributes.ControlTypeEnum.NumericUpDown:
                                AddNumericUpDown(setting.Name);
                                break;

                            case Attributes.ControlTypeEnum.ProgressBar:
                                AddProgressBar(setting.Name);
                                break;

                            case Attributes.ControlTypeEnum.Textbox:
                                AddTextBox(setting.Name);
                                break;

                            default:
                                AddDropDown(setting.Name);
                                break;
                        }
                    }
                    else
                    {
                        // If controlType was not passed then assume its a dropdown
                        AddDropDown(setting.Name);
                    }
                }
                catch (Exception e) {  }
            }
        }

        /// <summary>
        /// Provide the property name and the control will be added using all the Attributes in the Setting.cs class
        /// </summary>
        /// <param name="settingPropertyName"></param>
        public void AddDropDown(string settingPropertyName)
        {
            FPSDropdown dropdown = new FPSDropdown();
            object propValue = null;
            PropertyInfo pInfo = SettingType.GetProperties().FirstOrDefault(property => property.Name == settingPropertyName);

            if (pInfo == null)
            {
                Log.FailLog("=== ah shit it failed to get property info");
                return;
            }
            
            propValue = pInfo.GetValue(SettingInstance, null);

            // Enumerate all attributes associated with this property and pull out what we need
            AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[settingPropertyName].Attributes;
            DisplayNameAttribute displayName = (DisplayNameAttribute)attributes[typeof(DisplayNameAttribute)];
            DescriptionAttribute description = (DescriptionAttribute) attributes[typeof (DescriptionAttribute)];
            Attributes.SpellRequired spellRequired = (Attributes.SpellRequired) attributes[typeof (Attributes.SpellRequired)];
            Attributes.DropdownItems dropdownItems = (Attributes.DropdownItems) attributes[typeof (Attributes.DropdownItems)];

            if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name))
            {
                dropdown.SpellName = spellRequired.Name;
                dropdown.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);
            }

            string label = (displayName != null && !string.IsNullOrEmpty(displayName.DisplayName)) ? displayName.DisplayName : settingPropertyName.CamelToSpaced().Trim();

            if (dropdownItems != null && dropdownItems.Values.Any())
            {
                dropdown.Dropdown.Items.AddRange(dropdownItems.Values.ToArray());
            }
            else
            {
                // The DropdownItems attribute was omitted, so we will just add Always and Never
                dropdown.Dropdown.Items.AddRange(new object[] {"Always", "Never"});
            }

            if (description != null && !string.IsNullOrEmpty(description.Description))
            {
                dropdown.TooltipText = description.Description;
                dropdown.TooltipTitle = !string.IsNullOrEmpty(dropdown.SpellName) ? dropdown.SpellName : dropdown.Label;
            }

            dropdown.LabelSize = 130;
            dropdown.SettingName = settingPropertyName;
            dropdown.Label = label;
            dropdown.ValueChanged += FPSDropdown_ValueChangedEvent;
            dropdown.Value = propValue.ToString();

            panel1.Controls.Add(dropdown);
        }

        /// <summary>
        /// Provide the property name and the control will be added using all the Attributes in the Setting.cs class
        /// </summary>
        /// <param name="settingPropertyName"></param>
        public void AddProgressBar(string settingPropertyName)
        {
            FPSProgressBar bar = new FPSProgressBar();
            object propValue = null;
            PropertyInfo pInfo = SettingType.GetProperties().FirstOrDefault(property => property.Name == settingPropertyName);

            int minRange = 0;
            int maxRange = 100;
            Color barColor = Color.LimeGreen;
            
            if (pInfo == null)
            {
                Log.FailLog("=== ah shit it failed to get property info");
                return;
            }

            propValue = pInfo.GetValue(SettingInstance, null);

            // Enumerate all attributes associated with this property and pull out what we need
            AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[settingPropertyName].Attributes;
            DisplayNameAttribute displayName = (DisplayNameAttribute)attributes[typeof(DisplayNameAttribute)];
            DescriptionAttribute description = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
            Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];
            Attributes.MinMaxRange minMaxRange = (Attributes.MinMaxRange)attributes[typeof(Attributes.MinMaxRange)];
            Attributes.PowerType powerType = (Attributes.PowerType)attributes[typeof(Attributes.PowerType)];

            if (minMaxRange != null)
            {
                minRange = minMaxRange.MinimumRange;
                maxRange = minMaxRange.MaximumRange;
            }

            if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name))
            {
                bar.SpellName = spellRequired.Name;
                bar.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);
            }

            string label = (displayName != null && !string.IsNullOrEmpty(displayName.DisplayName)) ? displayName.DisplayName : settingPropertyName.CamelToSpaced().Trim();

            if (description != null && !string.IsNullOrEmpty(description.Description))
            {
                bar.TooltipText = description.Description;
                bar.TooltipTitle = !string.IsNullOrEmpty(bar.SpellName) ? bar.SpellName : bar.Label;
            }

            if (powerType != null)
            {
                switch (powerType.Name)
                {
                    case Attributes.PowerTypeEnum.Mana:
                        barColor = Color.DeepSkyBlue;
                        break;

                    case Attributes.PowerTypeEnum.Rage:
                        barColor = Color.FromArgb(224, 8, 8);
                        break;

                    case Attributes.PowerTypeEnum.Energy:
                        barColor = Color.FromArgb(244, 208, 0);
                        break;

                    default:
                        barColor = Color.LimeGreen;
                        break;
                }
            }


            bar.BarColor = barColor;
            bar.MinValue = minRange;
            bar.MaxValue = maxRange;
            bar.LabelSize = 130;
            bar.SettingName = settingPropertyName;
            bar.Label = label;
            bar.ValueChanged += FPSProgressBar_ValueChangedEvent;
            bar.Value = (int)propValue;

            panel1.Controls.Add(bar);
        }

        /// <summary>
        /// Provide the property name and the control will be added using all the Attributes in the Setting.cs class
        /// </summary>
        /// <param name="settingPropertyName"></param>
        public void AddNumericUpDown(string settingPropertyName)
        {
            FPSNumericUpdown num = new FPSNumericUpdown();
            object propValue = null;
            PropertyInfo pInfo = SettingType.GetProperties().FirstOrDefault(property => property.Name == settingPropertyName);

            int minRange = 1;
            int maxRange = 100;

            if (pInfo == null)
            {
                Log.FailLog("=== ah shit it failed to get property info");
                return;
            }

            propValue = pInfo.GetValue(SettingInstance, null);

            // Enumerate all attributes associated with this property and pull out what we need
            AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[settingPropertyName].Attributes;
            DisplayNameAttribute displayName = (DisplayNameAttribute)attributes[typeof(DisplayNameAttribute)];
            DescriptionAttribute description = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
            Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];
            Attributes.MinMaxRange minMaxRange = (Attributes.MinMaxRange)attributes[typeof(Attributes.MinMaxRange)];

            if (minMaxRange != null)
            {
                minRange = minMaxRange.MinimumRange;
                maxRange = minMaxRange.MaximumRange;
            }

            if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name))
            {
                num.SpellName = spellRequired.Name;
                num.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);
            }

            string label = (displayName != null && !string.IsNullOrEmpty(displayName.DisplayName)) ? displayName.DisplayName : settingPropertyName.CamelToSpaced().Trim();

            if (description != null && !string.IsNullOrEmpty(description.Description))
            {
                num.TooltipText = description.Description;
                num.TooltipTitle = !string.IsNullOrEmpty(num.SpellName) ? num.SpellName : num.Label;
            }

            num.MinValue = minRange;
            num.MaxValue = maxRange;
            num.LabelSize = 130;
            num.SettingName = settingPropertyName;
            num.Label = label;
            num.ValueChanged += FPSNumericUpDown_ValueChangedEvent;
            num.Value = (int)propValue;

            panel1.Controls.Add(num);
        }

        /// <summary>
        /// Provide the property name and the control will be added using all the Attributes in the Setting.cs class
        /// </summary>
        /// <param name="settingPropertyName"></param>
        public void AddTextBox(string settingPropertyName)
        {
            FPSTextBox box = new FPSTextBox();
            object propValue = null;
            PropertyInfo pInfo = SettingType.GetProperties().FirstOrDefault(property => property.Name == settingPropertyName);

            if (pInfo == null)
            {
                Log.FailLog("=== ah shit it failed to get property info");
                return;
            }

            propValue = pInfo.GetValue(SettingInstance, null);

            // Enumerate all attributes associated with this property and pull out what we need
            AttributeCollection attributes = TypeDescriptor.GetProperties(SettingType)[settingPropertyName].Attributes;
            DisplayNameAttribute displayName = (DisplayNameAttribute)attributes[typeof(DisplayNameAttribute)];
            DescriptionAttribute description = (DescriptionAttribute)attributes[typeof(DescriptionAttribute)];
            Attributes.SpellRequired spellRequired = (Attributes.SpellRequired)attributes[typeof(Attributes.SpellRequired)];

            if (spellRequired != null && !string.IsNullOrEmpty(spellRequired.Name))
            {
                box.SpellName = spellRequired.Name;
                box.IsSpellValid = SpellManager.HasSpell(spellRequired.Name);
            }

            string label = (displayName != null && !string.IsNullOrEmpty(displayName.DisplayName)) ? displayName.DisplayName : settingPropertyName.CamelToSpaced().Trim();

            if (description != null && !string.IsNullOrEmpty(description.Description))
            {
                box.TooltipText = description.Description;
                box.TooltipTitle = !string.IsNullOrEmpty(box.SpellName) ? box.SpellName : box.Label;
            }

            box.LabelSize = 130;
            box.SettingName = settingPropertyName;
            box.Label = label;
            box.ValueChanged += FPSTextBox_ValueChangedEvent;
            box.Value = propValue.ToString();

            panel1.Controls.Add(box);
        }

        public void AddDropDown(string label, string settingName, List<string> options, string tooltipTitle, string tooltipText)
        {
            FPSDropdown uc = new FPSDropdown
            {
                Label = label,
                SettingName = settingName,
                TooltipTitle = tooltipTitle,
                TooltipText = tooltipText,
                LabelSize = 130,
            };

            foreach (string s in options) uc.DropDownItems.Add(s);

            panel1.Controls.Add(uc);
        }

        public void AddTextBox(string label, string settingName, string tooltipTitle, string tooltipText)
        {
            FPSTextBox uc = new FPSTextBox
            {
                Label = label,
                SettingName = settingName,
                TooltipTitle = tooltipTitle,
                TooltipText = tooltipText,
                LabelSize = 130,
            };

            panel1.Controls.Add(uc);
        }

        public void AddNumericUpdown(string label, string settingName, int minValue, int maxValue, string tooltipTitle, string tooltipText)
        {
            FPSNumericUpdown uc = new FPSNumericUpdown
            {
                Label = label,
                SettingName = settingName,
                TooltipTitle = tooltipTitle,
                TooltipText = tooltipText,
                MinValue = minValue,
                MaxValue = maxValue,
                LabelSize = 130,
            };

            panel1.Controls.Add(uc);
        }

        public void AddProgressBar(string label, string settingName, int minValue, int maxValue, string tooltipTitle, string tooltipText)
        {
            FPSProgressBar uc = new FPSProgressBar
            {
                Label = label,
                SettingName = settingName,
                MaxValue = maxValue,
                MinValue = minValue,
                TooltipTitle = tooltipTitle,
                TooltipText = tooltipText,
                LabelSize = 130,
            };

            panel1.Controls.Add(uc);
        }


        private const int _padding = 1;
        private int _newtop = 8;
        private const int _top = 8;
        private const int _left = 6;
        
        private void OrganiseAllControls()
        {
            _newtop = _top;

            foreach (Control c in panel1.Controls)
            {
                c.Location = new Point(_left, _newtop);
                c.Width = panel1.Width - (_left*2);
                _newtop += (c.Size.Height + _padding);
            }
        }

    }




    /// <summary>
    /// Wrapper for the Additional popup UI
    /// </summary>
    public class AdditionalsPopup
    {
        private readonly UIAdditionals _additionals;
        private readonly Popup.Popup _popup;
        private readonly Control _owner;

        public PopupAnimations AanimationType
        {
            get { return _popup.ShowingAnimation; }
            set { _popup.ShowingAnimation = value; }
        }

        public bool DropShadowEnabled
        {
            get { return _popup.DropShadowEnabled; }
            set { _popup.DropShadowEnabled = value; }
        }

        public AdditionalsPopup(Settings settingInstance, Type settingType, Control ownerControl)
        {
            _additionals = new UIAdditionals { SettingInstance = settingInstance, SettingType = settingType };
            _popup = new Popup.Popup(_additionals) { Resizable = false, DropShadowEnabled = true, ShowingAnimation = PopupAnimations.Center };
            _owner = ownerControl;
        }

        public AdditionalsPopup(Settings settingInstance, Type settingType, Control ownerControl, string addCategory)
        {
            _additionals = new UIAdditionals { SettingInstance = settingInstance, SettingType = settingType };
            _popup = new Popup.Popup(_additionals) { Resizable = false, DropShadowEnabled = true, ShowingAnimation = PopupAnimations.Center };
            _owner = ownerControl;
            _additionals.AddCategory(addCategory);
        }

        public void ShowPopup()
        {
            const int controlHeight = 25;
            const int topPadding = 10;
            const int betweenPadding = 5;
            const int footerPadding = 50;
            int countOfControls = _additionals.CountOfControls;

            int newHeight = ((countOfControls*(controlHeight)) + topPadding + footerPadding);// + (countOfControls * betweenPadding));
            _popup.Size = new Size(_popup.Width, newHeight);

            _popup.Show(_owner);
        }

        public void AddDropDown(string propertyName)
        {
            _additionals.AddDropDown(propertyName);
        }

        public void AddProgressBar(string propertyName)
        {
            _additionals.AddProgressBar(propertyName);
        }

        public void AddNumericUpDown(string propertyName)
        {
            _additionals.AddNumericUpDown(propertyName);
        }

        public void AddTextBox(string propertyName)
        {
            _additionals.AddTextBox(propertyName);
        }

        /// <summary>
        /// Add all properties within the specified category
        /// </summary>
        /// <param name="propertyName"></param>
        public void AddCategory(string propertyName)
        {
            _additionals.AddCategory(propertyName);
        }
    }
}
